import pytest
import numpy as np
import pandas as pd

from orbit.models import KTRLite
from orbit.diagnostics.metrics import smape

# used for in-sample training insanity check
SMAPE_TOLERANCE = 0.28


@pytest.mark.parametrize(
    "seasonality_fs_order", [None, [5]], ids=["default_order", "manual_order"]
)
@pytest.mark.parametrize(
    "make_daily_data", [({"seasonality": "single", "with_coef": False})], indirect=True
)
def test_ktrlite_single_seas(make_daily_data, seasonality_fs_order):
    train_df, _, _ = make_daily_data

    ktrlite = KTRLite(
        response_col="response",
        date_col="date",
        seasonality=[365.25],
        seasonality_fs_order=seasonality_fs_order,
        estimator="stan-map",
        n_bootstrap_draws=-1,
    )

    ktrlite.fit(train_df)
    predict_df = ktrlite.predict(train_df)

    expected_columns = ["date", "prediction"]
    expected_shape = (train_df.shape[0], len(expected_columns))
    expected_num_parameters = len(ktrlite._model.get_model_param_names()) + 1

    assert predict_df.shape == expected_shape
    assert predict_df.columns.tolist() == expected_columns
    assert len(ktrlite._posterior_samples) == expected_num_parameters
    smape_val = smape(train_df["response"].values, predict_df["prediction"].values)
    assert smape_val <= SMAPE_TOLERANCE


@pytest.mark.parametrize(
    "seasonality_fs_order", [None, [2, 5]], ids=["default_order", "manual_order"]
)
@pytest.mark.parametrize(
    "make_daily_data", [({"with_dual_sea": True, "with_coef": False})], indirect=True
)
def test_ktrlite_dual_seas(make_daily_data, seasonality_fs_order):
    train_df, _, _ = make_daily_data

    ktrlite = KTRLite(
        response_col="response",
        date_col="date",
        seasonality=[7, 365.25],
        seasonality_fs_order=seasonality_fs_order,
        estimator="stan-map",
        n_bootstrap_draws=-1,
    )

    ktrlite.fit(train_df)
    predict_df = ktrlite.predict(train_df)

    expected_columns = ["date", "prediction"]
    expected_shape = (train_df.shape[0], len(expected_columns))
    expected_num_parameters = len(ktrlite._model.get_model_param_names()) + 1

    assert predict_df.shape == expected_shape
    assert predict_df.columns.tolist() == expected_columns
    assert len(ktrlite._posterior_samples) == expected_num_parameters
    smape_val = smape(train_df["response"].values, predict_df["prediction"].values)
    assert smape_val <= SMAPE_TOLERANCE


@pytest.mark.parametrize(
    "make_daily_data", [({"with_dual_sea": True, "with_coef": False})], indirect=True
)
@pytest.mark.parametrize("level_segments", [20, 10, 2])
def test_ktrlite_level_segments(make_daily_data, level_segments):
    train_df, test_df, coef = make_daily_data

    ktrlite = KTRLite(
        response_col="response",
        date_col="date",
        level_segments=level_segments,
        estimator="stan-map",
        n_bootstrap_draws=-1,
    )

    ktrlite.fit(train_df)
    predict_df = ktrlite.predict(test_df)

    expected_columns = ["date", "prediction"]
    expected_shape = (364, len(expected_columns))
    expected_num_parameters = len(ktrlite._model.get_model_param_names()) + 1

    assert predict_df.shape == expected_shape
    assert predict_df.columns.tolist() == expected_columns
    assert len(ktrlite._posterior_samples) == expected_num_parameters
    knots_df = ktrlite.get_level_knots()
    levels_df = ktrlite.get_levels()
    assert knots_df.shape[0] in [level_segments + 1, level_segments + 2]
    assert levels_df.shape[0] == ktrlite.get_training_meta()["num_of_obs"]


@pytest.mark.parametrize(
    "level_knot_dates",
    [
        pd.date_range(start="2016-03-01", end="2019-01-01", freq="3M"),
        pd.date_range(start="2016-03-01", end="2019-01-01", freq="6M"),
    ],
)
@pytest.mark.parametrize(
    "make_daily_data", [({"seasonality": "single", "with_coef": False})], indirect=True
)
def test_ktrlite_level_knot_dates(make_daily_data, level_knot_dates):
    train_df, test_df, coef = make_daily_data

    ktrlite = KTRLite(
        response_col="response",
        date_col="date",
        level_knot_dates=level_knot_dates,
        estimator="stan-map",
        n_bootstrap_draws=1e4,
    )

    ktrlite.fit(train_df)
    predict_df = ktrlite.predict(test_df)

    expected_columns = ["date", "prediction_5", "prediction", "prediction_95"]
    expected_shape = (364, len(expected_columns))
    expected_num_parameters = len(ktrlite._model.get_model_param_names()) + 1

    assert predict_df.shape == expected_shape
    assert predict_df.columns.tolist() == expected_columns
    assert len(ktrlite._posterior_samples) == expected_num_parameters
    assert np.all(np.isin(ktrlite._model.level_knot_dates, level_knot_dates))
    assert len(ktrlite._model.level_knot_dates) == len(level_knot_dates)


@pytest.mark.parametrize("level_knot_distance", [90, 120])
@pytest.mark.parametrize(
    "make_daily_data", [({"seasonality": "single", "with_coef": False})], indirect=True
)
def test_ktrlite_level_knot_distance(make_daily_data, level_knot_distance):
    train_df, test_df, coef = make_daily_data

    ktrlite = KTRLite(
        response_col="response",
        date_col="date",
        level_knot_distance=level_knot_distance,
        estimator="stan-map",
        n_bootstrap_draws=1e4,
    )

    ktrlite.fit(train_df)
    predict_df = ktrlite.predict(test_df)

    expected_columns = ["date", "prediction_5", "prediction", "prediction_95"]
    expected_shape = (364, len(expected_columns))
    expected_num_parameters = len(ktrlite._model.get_model_param_names()) + 1

    assert predict_df.shape == expected_shape
    assert predict_df.columns.tolist() == expected_columns
    assert len(ktrlite._posterior_samples) == expected_num_parameters


@pytest.mark.parametrize(
    "seas_segments",
    [
        pytest.param(0, id="0-seas_segement"),
        pytest.param(1, id="1-seas_segement"),
        pytest.param(5, id="5-seas_segement"),
    ],
)
@pytest.mark.parametrize(
    "make_daily_data", [({"seasonality": "single", "with_coef": False})], indirect=True
)
def test_ktrlite_seas_segments(make_daily_data, seas_segments):
    train_df, test_df, coef = make_daily_data

    ktrlite = KTRLite(
        response_col="response",
        date_col="date",
        seasonality=[7, 365.25],
        seasonality_fs_order=[2, 5],
        level_segments=10,
        seasonality_segments=seas_segments,
        estimator="stan-map",
        n_bootstrap_draws=-1,
    )

    ktrlite.fit(train_df)
    predict_df = ktrlite.predict(test_df)

    expected_columns = ["date", "prediction"]
    expected_shape = (364, len(expected_columns))
    expected_num_parameters = len(ktrlite._model.get_model_param_names()) + 1

    assert predict_df.shape == expected_shape
    assert predict_df.columns.tolist() == expected_columns
    assert len(ktrlite._posterior_samples) == expected_num_parameters


@pytest.mark.parametrize(
    "make_daily_data", [({"seasonality": "single", "with_coef": False})], indirect=True
)
def test_ktrlite_predict_decompose(make_daily_data):
    train_df, test_df, coef = make_daily_data

    ktrlite = KTRLite(
        response_col="response",
        date_col="date",
        seasonality=[7, 365.25],
        seasonality_fs_order=[2, 5],
        estimator="stan-map",
        n_bootstrap_draws=1e4,
    )

    ktrlite.fit(train_df)
    predict_df = ktrlite.predict(test_df, decompose=True)

    expected_columns = [
        "date",
        "prediction_5",
        "prediction",
        "prediction_95",
        "trend_5",
        "trend",
        "trend_95",
        "seasonality_7_5",
        "seasonality_7",
        "seasonality_7_95",
        "seasonality_365.25_5",
        "seasonality_365.25",
        "seasonality_365.25_95",
    ]
    expected_shape = (364, len(expected_columns))
    expected_num_parameters = len(ktrlite._model.get_model_param_names()) + 1

    assert predict_df.shape == expected_shape
    assert predict_df.columns.tolist() == expected_columns
    assert len(ktrlite._posterior_samples) == expected_num_parameters


@pytest.mark.parametrize(
    "make_daily_data", [({"seasonality": "single", "with_coef": False})], indirect=True
)
def test_ktrlite_predict_decompose_point_estimate(make_daily_data):
    train_df, test_df, coef = make_daily_data

    ktrlite = KTRLite(
        response_col="response",
        date_col="date",
        seasonality=[7, 365.25],
        seasonality_fs_order=[2, 5],
        estimator="stan-map",
        n_bootstrap_draws=-1,
    )

    ktrlite.fit(train_df)
    predict_df = ktrlite.predict(test_df, decompose=True)

    expected_columns = [
        "date",
        "prediction",
        "trend",
        "seasonality_7",
        "seasonality_365.25",
    ]
    expected_shape = (364, len(expected_columns))
    expected_num_parameters = len(ktrlite._model.get_model_param_names()) + 1

    assert predict_df.shape == expected_shape
    assert predict_df.columns.tolist() == expected_columns
    assert len(ktrlite._posterior_samples) == expected_num_parameters


def test_ktrlite_hourly_data(ca_hourly_electricity_data):
    train_df = ca_hourly_electricity_data

    ktrlite = KTRLite(
        response_col="SDGE",
        date_col="Dates",
        seasonality=[24, 7, 365.25],
        seasonality_fs_order=[3, 3, 5],
        estimator="stan-map",
        n_bootstrap_draws=-1,
    )

    ktrlite.fit(train_df)
    predict_df = ktrlite.predict(train_df)

    expected_columns = ["Dates", "prediction"]
    expected_shape = (train_df.shape[0], len(expected_columns))
    expected_num_parameters = len(ktrlite._model.get_model_param_names()) + 1

    assert predict_df.shape == expected_shape
    assert predict_df.columns.tolist() == expected_columns
    assert len(ktrlite._posterior_samples) == expected_num_parameters
    smape_val = smape(train_df["SDGE"].values, predict_df["prediction"].values)
    assert smape_val <= SMAPE_TOLERANCE



# from inspect import signature
from ..template.lgt import LGTModel
from ..forecaster import MAPForecaster, FullBayesianForecaster, SVIForecaster
from ..exceptions import IllegalArgument
from ..constants.constants import EstimatorsKeys


def LGT(
    seasonality=None,
    seasonality_sm_input=None,
    level_sm_input=None,
    regressor_col=None,
    regressor_sign=None,
    regressor_beta_prior=None,
    regressor_sigma_prior=None,
    regression_penalty="fixed_ridge",
    lasso_scale=0.5,
    auto_ridge_scale=0.5,
    slope_sm_input=None,
    estimator="stan-mcmc",
    suppress_stan_log=True,
    **kwargs,
):
    """
    Args
    ----------
    seasonality : int
        Length of seasonality
    seasonality_sm_input : float
        float value between [0, 1], applicable only if `seasonality` > 1. A larger value puts
        more weight on the current seasonality.
        If None, the model will estimate this value.
    level_sm_input : float
        float value between [0.0001, 1]. A larger value puts more weight on the current level.
        If None, the model will estimate this value.
    regressor_col : list
        Names of regressor columns, if any
    regressor_sign :  list
        list with values { '+', '-', '=' } such that
        '+' indicates regressor coefficient estimates are constrained to [0, inf).
        '-' indicates regressor coefficient estimates are constrained to (-inf, 0].
        '=' indicates regressor coefficient estimates can be any value between (-inf, inf).
        The length of `regressor_sign` must be the same length as `regressor_col`. If None,
        all elements of list will be set to '='.
    regressor_beta_prior : list
        list of prior float values for regressor coefficient betas. The length of `regressor_beta_prior`
        must be the same length as `regressor_col`. If None, use non-informative priors.
    regressor_sigma_prior : list
        list of prior float values for regressor coefficient sigmas. The length of `regressor_sigma_prior`
        must be the same length as `regressor_col`. If None, use non-informative priors.
    regression_penalty : { 'fixed_ridge', 'lasso', 'auto_ridge' }
        regression penalty method
    lasso_scale : float
        float value between [0, 1], applicable only if `regression_penalty` == 'lasso'
    auto_ridge_scale : float
        float value between [0, 1], applicable only if `regression_penalty` == 'auto_ridge'
    slope_sm_input : float
        float value between [0, 1]. A larger value puts more weight on the current slope.
        If None, the model will estimate this value.
    estimator : string; {'stan-mcmc', 'stan-map', 'pyro-svi'}
        default to be 'stan-mcmc'.

    Other Parameters
    ----------------
    response_col : str
        Name of response variable column, default 'y'
    date_col : str
        Name of date variable column, default 'ds'
    n_bootstrap_draws : int
        Number of samples to bootstrap in order to generate the prediction interval. For full Bayesian and
        variational inference forecasters, samples are drawn directly from original posteriors. For point-estimated
        posteriors, it will be used to sample noise parameters.  When -1 or None supplied, full Bayesian and
        variational inference forecasters will assume number of draws equal the size of original samples while
        point-estimated posteriors will mute the draw and output prediction without interval.
    prediction_percentiles : list
        List of integers of prediction percentiles that should be returned on prediction. To avoid reporting any
        confident intervals, pass an empty list
    suppress_stan_log : bool
        If False, turn off cmdstanpy logger. Default as False.

    **kwargs:
        additional arguments passed into orbit.estimators.stan_estimator or orbit.estimators.pyro_estimator
    """
    _supported_estimators = [
        EstimatorsKeys.StanMAP.value,
        EstimatorsKeys.StanMCMC.value,
        EstimatorsKeys.PyroSVI.value,
    ]

    lgt = LGTModel(
        seasonality=seasonality,
        seasonality_sm_input=seasonality_sm_input,
        level_sm_input=level_sm_input,
        regressor_col=regressor_col,
        regressor_sign=regressor_sign,
        regressor_beta_prior=regressor_beta_prior,
        regressor_sigma_prior=regressor_sigma_prior,
        regression_penalty=regression_penalty,
        lasso_scale=lasso_scale,
        auto_ridge_scale=auto_ridge_scale,
        slope_sm_input=slope_sm_input,
    )
    if estimator == EstimatorsKeys.StanMAP.value:
        from ..estimators.stan_estimator import StanEstimatorMAP

        lgt_forecaster = MAPForecaster(
            model=lgt,
            estimator_type=StanEstimatorMAP,
            suppress_stan_log=suppress_stan_log,
            **kwargs,
        )
    elif estimator == EstimatorsKeys.StanMCMC.value:
        from ..estimators.stan_estimator import StanEstimatorMCMC

        lgt_forecaster = FullBayesianForecaster(
            model=lgt,
            estimator_type=StanEstimatorMCMC,
            suppress_stan_log=suppress_stan_log,
            **kwargs,
        )
    elif estimator == EstimatorsKeys.PyroSVI.value:
        from ..estimators.pyro_estimator import PyroEstimatorSVI

        lgt_forecaster = SVIForecaster(
            model=lgt, estimator_type=PyroEstimatorSVI, **kwargs
        )
    else:
        raise IllegalArgument(
            "Invalid estimator. Must be one of {}".format(_supported_estimators)
        )

    return lgt_forecaster



from ..template.ktr import KTRModel
from ..forecaster import SVIForecaster
from ..exceptions import IllegalArgument
from ..estimators.pyro_estimator import PyroEstimatorSVI
from ..constants.constants import EstimatorsKeys


def KTR(
    level_knot_scale=0.1,
    level_segments=10,
    level_knot_distance=None,
    level_knot_dates=None,
    # seasonality
    seasonality=None,
    seasonality_fs_order=None,
    seasonality_segments=3,
    seasonal_initial_knot_scale=1.0,
    seasonal_knot_scale=0.1,
    # regression
    regressor_col=None,
    regressor_sign=None,
    regressor_init_knot_loc=None,
    regressor_init_knot_scale=None,
    regressor_knot_scale=None,
    regression_segments=5,
    regression_knot_distance=None,
    regression_knot_dates=None,
    # different from seasonality
    regression_rho=0.15,
    # shared
    degree_of_freedom=30,
    date_freq=None,
    # time-based coefficient priors
    coef_prior_list=None,
    flat_multiplier=True,
    # TODO: rename to residuals upper bound
    residuals_scale_upper=None,
    ktrlite_optim_args=dict(),
    estimator="pyro-svi",
    **kwargs,
):
    """
    Parameters
    ----------
    level_knot_scale : float
        sigma for level; default to be .1
    level_segments : int
        the number of segments partitioned by the knots of level (trend)
    level_knot_distance : int
        the distance between every two knots of level (trend)
    level_knot_dates : array like
        list of pre-specified dates for the level knots
    seasonality : int, or list of int
        multiple seasonality
    seasonality_fs_order : int, or list of int
        fourier series order for seasonality
    seasonality_segments : int
        the number of segments partitioned by the knots of seasonality
    seasonal_initial_knot_scale : float
        scale parameter for seasonal regressors initial coefficient knots; default to be 1
    seasonal_knot_scale : float
        scale parameter for seasonal regressors drift of coefficient knots; default to be 0.1.
    regressor_col : array-like strings
        regressor columns
    regressor_sign : list
        list of signs with '=' for regular regressor and '+' for positive regressor
    regressor_init_knot_loc : list
        list of regressor knot pooling mean priors, default to be 0's
    regressor_init_knot_scale : list
        list of regressor knot pooling sigma's to control the pooling strength towards the grand mean of regressors;
        default to be 1.
    regressor_knot_scale : list
        list of regressor knot sigma priors; default to be 0.1.
    regression_segments : int
        the number of segments partitioned by the knots of regression
    regression_knot_distance : int
        the distance between every two knots of regression
    regression_knot_dates : array-like
        list of pre-specified dates for regression knots
    regression_rho : float
        sigma in the Gaussian kernel for the regression term
    degree_of_freedom : int
        degree of freedom for error t-distribution
    date_freq : str
        date frequency; if not supplied, pd.infer_freq will be used to imply the date frequency.
    coef_prior_list : list of dicts
        each dict in the list should have keys as
        'name', prior_start_tp_idx' (inclusive), 'prior_end_tp_idx' (not inclusive),
        'prior_mean', 'prior_sd', and 'prior_regressor_col'
    residuals_scale_upper : float
    flat_multiplier : bool
        Default set as True. If False, we will adjust knot scale with a multiplier based on regressor volume
        around each knot; When True, set all multiplier as 1
    ktrlite_optim_args : dict
        the optimizing config for the ktrlite model (to fit level/seasonality). Default to be dict().
    estimator : string; {'pyro-svi'}

    Other Parameters
    ----------------
    response_col : str
        Name of response variable column, default 'y'
    date_col : str
        Name of date variable column, default 'ds'
    n_bootstrap_draws : int
        Number of samples to bootstrap in order to generate the prediction interval. For full Bayesian and
        variational inference forecasters, samples are drawn directly from original posteriors. For point-estimated
        posteriors, it will be used to sample noise parameters.  When -1 or None supplied, full Bayesian and
        variational inference forecasters will assume number of draws equal the size of original samples while
        point-estimated posteriors will mute the draw and output prediction without interval.
    prediction_percentiles : list
        List of integers of prediction percentiles that should be returned on prediction. To avoid reporting any
        confident intervals, pass an empty list

    **kwargs:
        additional arguments passed into orbit.estimators.pyro_estimator
    """
    _supported_estimators = [EstimatorsKeys.PyroSVI.value]

    ktr = KTRModel(
        level_knot_scale=level_knot_scale,
        level_segments=level_segments,
        level_knot_distance=level_knot_distance,
        level_knot_dates=level_knot_dates,
        seasonality=seasonality,
        seasonality_fs_order=seasonality_fs_order,
        seasonality_segments=seasonality_segments,
        seasonal_initial_knot_scale=seasonal_initial_knot_scale,
        seasonal_knot_scale=seasonal_knot_scale,
        regressor_col=regressor_col,
        regressor_sign=regressor_sign,
        regressor_init_knot_loc=regressor_init_knot_loc,
        regressor_init_knot_scale=regressor_init_knot_scale,
        regressor_knot_scale=regressor_knot_scale,
        regression_segments=regression_segments,
        regression_knot_distance=regression_knot_distance,
        regression_knot_dates=regression_knot_dates,
        regression_rho=regression_rho,
        degree_of_freedom=degree_of_freedom,
        coef_prior_list=coef_prior_list,
        date_freq=date_freq,
        flat_multiplier=flat_multiplier,
        residuals_scale_upper=residuals_scale_upper,
        ktrlite_optim_args=ktrlite_optim_args,
    )
    if estimator == EstimatorsKeys.PyroSVI.value:
        ktr_forecaster = SVIForecaster(
            model=ktr, estimator_type=PyroEstimatorSVI, **kwargs
        )
    else:
        raise IllegalArgument(
            "Invalid estimator. Must be one of {}".format(_supported_estimators)
        )

    return ktr_forecaster



import os
import numpy as np
import pandas as pd
from copy import deepcopy
from itertools import product


def update_dict(original_dict, append_dict):
    new_dict = deepcopy(original_dict)
    if append_dict:
        for k, v in append_dict.items():
            new_dict[k] = v
    return new_dict


def is_ordered_datetime(array):
    """Returns True if array is ordered and non-repetitive"""
    return np.all(np.diff(array).astype(float) > 0)


def is_even_gap_datetime(array):
    """Returns True if array is evenly distributed"""
    if len(array) >= 3:
        return isinstance(pd.infer_freq(array), str)
    return True


def is_empty_dataframe(df):
    """A simple function to tell whether the passed in df is an empty dataframe or not.
    Parameters
    ----------
    df : pd.DataFrame
        given input dataframe

    Returns
    -------
    bool : True if df is none, or if df is an empty dataframe; False otherwise.
    """
    return df is None or (isinstance(df, pd.DataFrame) and df.empty)


def get_parent_path(current_file_path):
    """
    Parameters
    ----------
    current_file_path : str
        The given file path, should be an absolute path

    Returns:
    -------
        str : The parent path of give file path
    """

    return os.path.abspath(os.path.join(current_file_path, os.pardir))


def expand_grid(base):
    """Given a base key values span, expand them into a dataframe covering all combinations
    Parameters
    ----------
    base : dict
        dictionary with keys equal columns name and value equals key values

    Returns
    -------
    pd.DataFrame : dataframe generate based on user specified base
    """
    return pd.DataFrame([row for row in product(*base.values())], columns=base.keys())


def regenerate_base_df(df, time_col, key_col, val_cols=[], fill_na=None):
    """Given a dataframe, key column, time column and value column, re-generate multiple time-series to cover full range
    date-time with all the keys.  This can be a useful utils for working multiple time-series.

    Parameters
    ----------
    df : pd.DataFrame
    time_col : str
    key_col : str
    val_cols : List[str]; values column considered to be imputed
    fill_na : Optional[float]; values to fill when there are missing values of the row

    Returns
    -------

    """
    out = df.copy()
    unique_time = out[time_col].unique()
    unique_key = out[key_col].unique()
    new_df_base = expand_grid(
        {
            key_col: unique_key,
            time_col: unique_time,
        }
    )
    out = new_df_base.merge(out, how="left", on=[time_col, key_col])
    if not isinstance(val_cols, list):
        val_cols = list(val_cols)
    out = out[[time_col, key_col] + val_cols]
    if fill_na is not None:
        out[val_cols] = out[val_cols].fillna(fill_na)
    return out



# from inspect import signature
from ..template.ets import ETSModel
from ..forecaster import MAPForecaster, FullBayesianForecaster
from ..exceptions import IllegalArgument
from ..constants.constants import EstimatorsKeys


def ETS(
    seasonality=None,
    seasonality_sm_input=None,
    level_sm_input=None,
    estimator="stan-mcmc",
    suppress_stan_log=True,
    **kwargs,
):
    """
    Args
    ----------
    seasonality : int
        Length of seasonality
    seasonality_sm_input : float
        float value between [0, 1], applicable only if `seasonality` > 1. A larger value puts
        more weight on the current seasonality.
        If None, the model will estimate this value.
    level_sm_input : float
        float value between [0.0001, 1]. A larger value puts more weight on the current level.
        If None, the model will estimate this value.
    estimator : string; {'stan-mcmc', 'stan-map'}
        default to be 'stan-mcmc'.

    Other Parameters
    ----------------
    response_col : str
        Name of response variable column, default 'y'
    date_col : str
        Name of date variable column, default 'ds'
    n_bootstrap_draws : int
        Number of samples to bootstrap in order to generate the prediction interval. For full Bayesian and
        variational inference forecasters, samples are drawn directly from original posteriors. For point-estimated
        posteriors, it will be used to sample noise parameters.  When -1 or None supplied, full Bayesian and
        variational inference forecasters will assume number of draws equal the size of original samples while
        point-estimated posteriors will mute the draw and output prediction without interval.
    prediction_percentiles : list
        List of integers of prediction percentiles that should be returned on prediction. To avoid reporting any
        confident intervals, pass an empty list
    suppress_stan_log : bool
        If False, turn off cmdstanpy logger. Default as False.

    **kwargs:
        additional arguments passed into orbit.estimators.stan_estimator
    """
    _supported_estimators = [
        EstimatorsKeys.StanMAP.value,
        EstimatorsKeys.StanMCMC.value,
    ]

    ets = ETSModel(
        seasonality=seasonality,
        seasonality_sm_input=seasonality_sm_input,
        level_sm_input=level_sm_input,
    )
    if estimator == EstimatorsKeys.StanMAP.value:
        from ..estimators.stan_estimator import StanEstimatorMAP

        ets_forecaster = MAPForecaster(
            model=ets,
            estimator_type=StanEstimatorMAP,
            suppress_stan_log=suppress_stan_log,
            **kwargs,
        )
    elif estimator == EstimatorsKeys.StanMCMC.value:
        from ..estimators.stan_estimator import StanEstimatorMCMC

        ets_forecaster = FullBayesianForecaster(
            model=ets,
            estimator_type=StanEstimatorMCMC,
            suppress_stan_log=suppress_stan_log,
            **kwargs,
        )
    else:
        raise IllegalArgument(
            "Invalid estimator. Must be one of {}".format(_supported_estimators)
        )

    return ets_forecaster



import numpy as np
import pandas as pd
from statsmodels.tsa.arima_process import ArmaProcess
import math
from math import pi
from orbit.exceptions import IllegalArgument


def make_trend(
    series_len, method="rw", arma=[0.25, 0.6], rw_loc=0.0, rw_scale=0.1, seed=1
):
    """Module to generate time-series trend with different methods
    Parameters
    ----------
    series_len: int
        Total length of series
    method: str ['arma', 'rw']
        In case of `'rw'`, a simple random walk process will be used. For `'arma'`, we will use `statsmodels.api` to
        simulate a simple ARMA(1, 1) process
    arma: list
        List [arparams, maparams] of size 2 where used for arma(1) generating process
    rw_loc: float
        Location parameter of random walk generated by `np.random.normal()`
    rw_scale: float
        Scale parameter of random walk generated by `np.random.normal()`
    seed: int
        Seed passed into `np.random.default_rng()`
    Returns
    -------
    np.array-llike
        Simulated trend with length equals `series_len`

    Notes
    -----
        1. ARMA process: https://www.statsmodels.org/stable/generated/statsmodels.tsa.arima_process.ArmaProcess.html
    """
    # make trend
    if method == "rw":
        rw = np.random.default_rng(seed).normal(rw_loc, rw_scale, series_len)
        trend = np.cumsum(rw)
    elif method == "arma":
        arparams = np.array([arma[0]])
        maparams = np.array([arma[1]])
        # add zero-lag and negate
        ar = np.r_[1, -arparams]
        # add zero-lag
        ma = np.r_[1, maparams]
        arma_process = ArmaProcess(ar, ma)
        trend = arma_process.generate_sample(series_len)
    else:
        raise IllegalArgument("Invalid trend method.")

    return trend


def make_seasonality(
    series_len, seasonality, method="discrete", order=3, duration=1, scale=0.05, seed=1
):
    """Module to generate time-series seasonality with different methods
    series_len: int
        Total length of series
    seasonality: int
        For example, seasonality=52 would be a weekly series
    method: str ['discrete', 'fourier']
    order: int
        Used when method = 'fourier' ONLY. Fourier series order to generate seasonality.
    duration: int
        Used when method = 'discrete'. For example, seasonality=52 and duration=7 would be a daily
        series with a constant level on each week.
    scale: float
        Scale parameter of seasonality generation by Normal(0, scale). When method = 'discrete', it is directly used.
        when method = 'fourier', it is used to multiply with the sin-cosine wave generated series.
    seed: int
        Seed passed into `np.random.default_rng()`
    Returns
    -------
    np.array-llike
        Simulated seasonliaty with length equals `series_len`

    Notes
    -----
      1. In case of method = 'discrete', only `seasonality - 1` variables will be generated and the rest is
      directly derived to preserve the property of the sum of seasonality equals 1.
      2.  In case of method = 'fourier', see https://otexts.com/fpp2/complexseasonality.html
    """
    if seasonality > 1:
        if method == "fourier":
            t = np.arange(0, series_len)
            out = []
            for i in range(1, order + 1):
                x = 2.0 * i * np.pi * t / seasonality
                out.append(np.sin(x))
                out.append(np.cos(x))
            out = np.column_stack(out)
            b = np.random.default_rng(seed).normal(0, scale, order * 2)
            seas = np.matmul(out, b)
        else:
            # initialization
            seas = []
            iterations = math.ceil(series_len / duration)
            # initialize vector to be repeated
            init_seas = np.zeros(seasonality)
            init_seas[:-1] = np.random.default_rng(seed).normal(
                0, scale, seasonality - 1
            )
            init_seas[seasonality - 1] = -1 * np.sum(init_seas)
            for idx in range(iterations):
                seas += [init_seas[idx % seasonality]] * duration
            seas = np.array(seas[:series_len])
    else:
        seas = np.zeros(series_len)
    return seas


def make_regression(
    series_len,
    coefs,
    loc=0.0,
    scale=0.5,
    cov=None,
    noise_scale=1.0,
    bias=None,
    relevance=1.0,
    sparsity=0.2,
    seed=1,
):
    """Module to generate multiplicative time-series with trend, seasonality and regression components
    Parameters
    ----------
    series_len: int
        Total length of series
    coefs: 1-D array_like
        Values used as regression coefficients
    loc: float
        Location parameter to generate regressor values
    scale: float
        Scale parameter to generate regressor values
    cov: 2-D array_like, of shape (num_of_regressors, num_of_regressors)
        Covariance of regressors; need to be positive definite
    noise_scale:
        Scale parameter in the white noise generation process
    bias: float
        bias or intercept of the regression component; if None, bias is set to zero.
    relevance: float
        0 to 1; smaller value indicates less number of total useful regressors
    sparsity: float
        0 to 1 to control probability (= 1 - sparsity) at time t of a regressor value > 0
    seed: int
        Seed passed into `np.random.default_rng()`

    Returns
    -------
    x: 2-D array like
        Regressors simulated array. Should be with shape (series_len, num_of_regressors);
    y: 1-D array like
        Regression derived by product of X and coefficients plus noise
    coefs: float
        Coefficients modified in the process due to sparsity; if sparsity=0, it should be identical to coefs inputted
        by user
    """

    num_of_regressors = len(coefs)
    if (relevance > 0.0) and (relevance < 1.0):
        num_of_irr_coefs = int(num_of_regressors * (1 - relevance))
        coefs = np.copy(coefs)
        irr_idx = np.random.choice(num_of_regressors, num_of_irr_coefs, replace=False)
        coefs[irr_idx] = 0.0

    if cov is None:
        x = (
            np.random.default_rng(seed)
            .normal(loc, scale, series_len * num_of_regressors)
            .reshape(series_len, -1)
        )
    else:
        x = np.random.default_rng(seed).multivariate_normal(
            np.array([loc] * num_of_regressors, dtype=np.float64), cov, series_len
        )
    # control probability of regression kick-in
    if (sparsity > 0.0) and (sparsity < 1.0):
        z = (
            np.random.default_rng(seed)
            .binomial(1, 1 - sparsity, series_len * num_of_regressors)
            .reshape(series_len, -1)
        )
        x = x * z

    noise = np.random.default_rng(seed).normal(0, noise_scale, series_len)
    # make observed response
    if bias:
        y = bias + np.matmul(x, coefs) + noise
    else:
        y = np.matmul(x, coefs) + noise

    return x, y, coefs


# def sim_stepwise_coef_data(n, RS, p=3, n_jump=2):
#     """
#     Parameters
#     ----------
#     n : int
#         number of obseravtions
#     RS : int
#         seed
#     p : int
#         number of predictors
#     n_jump
#         number of jumps
#     Returns
#     -------
#     np.ndarray
#
#     Notes
#     -----
#     This code is purely experimental
#
#     """
#     np.random.seed(RS)
#
#     # initializing coefficients at zeros, simulate all coefficient values
#     lev = np.cumsum(np.concatenate((np.array([5.0]), np.random.normal(0, 0.01, n - 1))))
#
#     jump_tp = np.round(np.random.uniform(0, n, (p, n_jump)), 0).astype(np.int32)
#     jump_amp = np.random.normal(0, 0.05, (p, n_jump))
#
#     beta = np.zeros((n, p))
#     beta[0, :] = -2
#     for idx in range(p):
#         beta[jump_tp[idx], idx] = jump_amp[idx]
#     beta = np.cumsum(beta, 0)
#     #     beta += np.random.normal(0, 1e-2, (n, p))
#     # strictly positive coefficients
#     beta = np.exp(beta)
#
#     # simulate regressors
#     covariates = np.random.normal(0, 10, (n, p))
#
#     # observation with noise
#     y = lev + (covariates * beta).sum(-1) + 0.3 * np.random.normal(0, 1, n)
#
#     regressor_col = ["x{}".format(pp) for pp in range(1, p + 1)]
#     data = pd.DataFrame(covariates, columns=regressor_col)
#     data["y"] = y
#     data["date"] = pd.date_range(start="1/1/2018", periods=len(y))
#
#     # hack for p = 3
#     data["beta1"] = beta[:, 0]
#     data["beta2"] = beta[:, 1]
#     data["beta3"] = beta[:, 2]
#
#     return data


# def sim_data_grw(n, RS, p=3):
#     """coefficients curve are geometric random walk like
#
#     Notes
#     -----
#     This code is purely experimental
#     """
#     np.random.seed(RS)
#     beta_init = np.concatenate(
#         (
#             np.random.normal(2.0, 1.0, size=(1, 1)),  # leves
#             np.random.normal(-3.0, 0.05, size=(1, p)),  # regression coefficients
#         ),
#         axis=-1,
#     )
#
#     beta_drift = np.random.normal(0.0, 0.05, size=(n - 1, p + 1))  # drift
#
#     beta = np.concatenate((beta_init, beta_drift), axis=0)
#
#     # geometric random walk
#     beta = np.exp(np.cumsum(beta, 0))
#
#     # simulate regressors
#     covar_lev = np.ones((n, 1))
#     covar = np.concatenate((covar_lev, np.random.normal(0, 10.0, (n, p))), axis=1)
#
#     # observation with noise
#     y = (covar * beta).sum(-1) + 0.3 * np.random.normal(0, 1, n)
#
#     regressor_col = ["x{}".format(pp) for pp in range(1, p + 1)]
#     data = pd.DataFrame(covar[:, 1:], columns=regressor_col)
#     beta_col = ["beta{}".format(pp) for pp in range(1, p + 1)]
#     beta_data = pd.DataFrame(beta[:, 1:], columns=beta_col)
#     data = pd.concat([data, beta_data], axis=1)
#
#     data["y"] = y
#     data["date"] = pd.date_range(start="1/1/2018", periods=len(y))
#
#     return data
#
#
# def sim_data_rw(n, RS, p=3):
#     """coefficients curve are random walk like"""
#     np.random.seed(RS)
#
#     # initializing coefficients at zeros, simulate all coefficient values
#     lev = np.cumsum(np.concatenate((np.array([5.0]), np.random.normal(0, 0.01, n - 1))))
#     beta = np.concatenate(
#         [
#             np.random.uniform(0.05, 0.12, size=(1, p)),
#             np.random.normal(0.0, 0.01, size=(n - 1, p)),
#         ],
#         axis=0,
#     )
#     beta = np.cumsum(beta, 0)
#
#     # simulate regressors
#     covariates = np.random.normal(0, 10, (n, p))
#
#     # observation with noise
#     y = lev + (covariates * beta).sum(-1) + 0.3 * np.random.normal(0, 1, n)
#
#     regressor_col = ["x{}".format(pp) for pp in range(1, p + 1)]
#     data = pd.DataFrame(covariates, columns=regressor_col)
#     beta_col = ["beta{}".format(pp) for pp in range(1, p + 1)]
#     beta_data = pd.DataFrame(beta, columns=beta_col)
#     data = pd.concat([data, beta_data], axis=1)
#
#     data["y"] = y
#     data["date"] = pd.date_range(start="1/1/2018", periods=len(y))
#
#     return data
#
#
# def sim_data_seasonal(n, RS):
#     """coefficients curve are sine-cosine like"""
#     np.random.seed(RS)
#     # make the time varying coefs
#     tau = np.arange(1, n + 1) / n
#     data = pd.DataFrame(
#         {
#             "tau": tau,
#             "date": pd.date_range(start="1/1/2018", periods=n),
#             "beta1": 2 * tau,
#             "beta2": 1.01 + np.sin(2 * pi * tau),
#             "beta3": 1.01 + np.sin(4 * pi * (tau - 1 / 8)),
#             "x1": np.random.normal(0, 10, size=n),
#             "x2": np.random.normal(0, 10, size=n),
#             "x3": np.random.normal(0, 10, size=n),
#             "trend": np.cumsum(
#                 np.concatenate((np.array([1]), np.random.normal(0, 0.1, n - 1)))
#             ),
#             "error": np.random.normal(0, 1, size=n),  # stats.t.rvs(30, size=n),#
#         }
#     )
#
#     data["y"] = (
#         data.x1 * data.beta1 + data.x2 * data.beta2 + data.x3 * data.beta3 + data.error
#     )
#     return data



import pytest

from orbit.models import DLT, LGT
from orbit.utils.params_tuning import grid_search_orbit


@pytest.mark.parametrize(
    "param_grid",
    [
        {
            "level_sm_input": [0.3, 0.5, 0.8],
            "seasonality_sm_input": [0.3, 0.5, 0.8],
        },
        {
            "damped_factor": [0.3, 0.5, 0.8],
            "slope_sm_input": [0.3, 0.5, 0.8],
        },
    ],
)
@pytest.mark.parametrize("eval_method", ["backtest", "bic"])
def test_dlt_grid_tuning(make_weekly_data, param_grid, eval_method):
    train_df, test_df, coef = make_weekly_data
    args = {
        "response_col": "response",
        "date_col": "week",
        "seasonality": 52,
        "estimator": "stan-map",
    }
    dlt = DLT(**args)

    best_params, tuned_df = grid_search_orbit(
        param_grid,
        model=dlt,
        df=train_df,
        eval_method=eval_method,
        min_train_len=80,
        incremental_len=20,
        forecast_len=20,
    )

    assert best_params[0].keys() == param_grid.keys()
    assert set(tuned_df.columns.to_list()) == set(list(param_grid.keys()) + ["metrics"])
    assert tuned_df.shape == (9, 3)


@pytest.mark.parametrize(
    "param_grid",
    [
        {
            "level_sm_input": [0.3, 0.5, 0.8],
            "seasonality_sm_input": [0.3, 0.5, 0.8],
        },
        {
            "level_sm_input": [0.3, 0.5, 0.8],
            "slope_sm_input": [0.3, 0.5, 0.8],
        },
    ],
)
@pytest.mark.parametrize("eval_method", ["backtest", "bic"])
def test_lgt_grid_tuning(make_weekly_data, param_grid, eval_method):
    train_df, test_df, coef = make_weekly_data
    args = {
        "response_col": "response",
        "date_col": "week",
        "seasonality": 52,
        "estimator": "stan-map",
    }
    lgt = LGT(**args)

    best_params, tuned_df = grid_search_orbit(
        param_grid,
        model=lgt,
        df=train_df,
        eval_method=eval_method,
        min_train_len=80,
        incremental_len=20,
        forecast_len=20,
    )

    assert best_params[0].keys() == param_grid.keys()
    assert set(tuned_df.columns.to_list()) == set(list(param_grid.keys()) + ["metrics"])
    assert tuned_df.shape == (9, 3)



import numpy as np
import pandas as pd
from functools import partial

from ..constants.constants import PredictMethod, PredictionKeys
from ..exceptions import ForecasterException
from ..utils.predictions import prepend_date_column, compute_percentiles
from .forecaster import Forecaster


class MAPForecaster(Forecaster):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._point_posteriors[PredictMethod.MAP.value] = dict()
        self._point_method = PredictMethod.MAP.value

    def set_forecaster_training_meta(self, data_input):
        # MCMC flag to be true
        data_input.update({"WITH_MCMC": 0})
        return data_input

    def fit(self, df, **kwargs):
        super().fit(df, **kwargs)
        posterior_samples = self._posterior_samples
        map_posterior = {}
        for param_name in self._model.get_model_param_names():
            param_array = posterior_samples[param_name]
            # add one dimension as batch to have consistent logic with `.predict()`
            param_array = np.expand_dims(param_array, axis=0)
            map_posterior.update({param_name: param_array})

        self._point_posteriors[PredictMethod.MAP.value] = map_posterior
        # TODO: right now this is hacky:
        #  need to do it one more time to over-write the extra methods with right posterior
        self.load_extra_methods()

        return self

    def predict(
        self, df, decompose=False, store_prediction_array=False, seed=None, **kwargs
    ) -> pd.DataFrame:
        # raise if model is not fitted
        if not self.is_fitted():
            raise ForecasterException("Model is not fitted yet.")
        # obtain basic meta data from input df
        self._set_prediction_meta(df)
        prediction_meta = self.get_prediction_meta()
        training_meta = self.get_training_meta()

        if seed is not None:
            np.random.seed(seed)

        # perform point prediction
        point_posteriors = self._point_posteriors.get(PredictMethod.MAP.value)
        point_predicted_dict = self._model.predict(
            posterior_estimates=point_posteriors,
            df=df,
            training_meta=training_meta,
            prediction_meta=prediction_meta,
            # false for point estimate
            include_error=False,
            **kwargs,
        )
        for k, v in point_predicted_dict.items():
            point_predicted_dict[k] = np.squeeze(v, 0)

        # to derive confidence interval; the condition should be sufficient since we add [50] by default
        if self._n_bootstrap_draws > 0 and len(self._prediction_percentiles) > 1:
            # perform bootstrap; we don't have posterior samples. hence, we just repeat the draw here.
            posterior_samples = {}
            for k, v in point_posteriors.items():
                posterior_samples[k] = np.repeat(v, self._n_bootstrap_draws, axis=0)
            predicted_dict = self._model.predict(
                posterior_estimates=posterior_samples,
                df=df,
                training_meta=training_meta,
                prediction_meta=prediction_meta,
                include_error=True,
                **kwargs,
            )
            if store_prediction_array:
                self.prediction_array = predicted_dict[PredictionKeys.PREDICTION.value]
            percentiles_dict = compute_percentiles(
                predicted_dict, self._prediction_percentiles
            )
            # replace mid point prediction by point estimate
            percentiles_dict.update(point_predicted_dict)

            if PredictionKeys.PREDICTION.value not in percentiles_dict.keys():
                raise ForecasterException(
                    "cannot find the key:'{}' from return of _predict()".format(
                        PredictionKeys.PREDICTION.value
                    )
                )

            # since we always assume to have decompose from .predict() at first,
            # here it reduces to prediction when decompose is not requested
            if not decompose:
                k = PredictionKeys.PREDICTION.value
                reduced_keys = [
                    k + "_" + str(p) if p != 50 else k
                    for p in self._prediction_percentiles
                ]
                percentiles_dict = {
                    k: v for k, v in percentiles_dict.items() if k in reduced_keys
                }
            predicted_df = pd.DataFrame(percentiles_dict)
        else:
            if not decompose:
                # reduce to prediction only if decompose is not requested
                point_predicted_dict = {
                    k: v
                    for k, v in point_predicted_dict.items()
                    if k == PredictionKeys.PREDICTION.value
                }
            predicted_df = pd.DataFrame(point_predicted_dict)

        predicted_df = prepend_date_column(predicted_df, df, self.date_col)
        return predicted_df

    # TODO: should be private
    def load_extra_methods(self):
        for method in self.extra_methods:
            setattr(
                self,
                method,
                partial(
                    getattr(self._model, method),
                    self.get_training_meta(),
                    PredictMethod.MAP.value,
                    self.get_point_posteriors(),
                    self.get_posterior_samples(),
                ),
            )

    def get_bic(self):
        training_metrics = self.get_training_metrics()
        loglk = training_metrics["loglk"]
        n = loglk.shape[0] * loglk.shape[1]
        k = training_metrics["num_of_params"]
        return -2.0 * np.sum(loglk) + k * np.log(n)



import pytest
import numpy as np
import pandas as pd
from orbit.utils.knots import get_knot_idx, get_knot_dates


@pytest.mark.parametrize("num_of_segments", [0, 1, 3, 10])
@pytest.mark.parametrize(
    "dataset, date_col",
    [
        ("iclaims_training_data", "week"),
        ("m3_monthly_data", "date"),
        ("ca_hourly_electricity_data", "Dates"),
    ],
)
def test_segments_args(dataset, date_col, num_of_segments, request):
    df = request.getfixturevalue(dataset)
    date_array = df[date_col]
    knot_idx = get_knot_idx(num_of_obs=df.shape[0], num_of_segments=num_of_segments)
    assert knot_idx[0] == 0
    assert len(knot_idx) == num_of_segments + 1
    freq = pd.infer_freq(date_array)
    knot_dates = get_knot_dates(date_array[0], knot_idx, freq)
    # first knot always at the first date
    assert knot_dates[0] == date_array[0]


@pytest.mark.parametrize("knot_distance", [2, 4])
@pytest.mark.parametrize(
    "dataset, date_col",
    [
        ("iclaims_training_data", "week"),
        ("m3_monthly_data", "date"),
        ("ca_hourly_electricity_data", "Dates"),
    ],
)
def test_distance_args(dataset, date_col, knot_distance, request):
    df = request.getfixturevalue(dataset)
    date_array = df[date_col]
    knot_idx = get_knot_idx(num_of_obs=df.shape[0], knot_distance=knot_distance)

    assert knot_idx[0] == 0
    assert knot_idx[3] - knot_idx[2] == knot_distance


@pytest.mark.parametrize(
    "dataset, date_col, knot_dates, knot_idx",
    [
        (
            "iclaims_training_data",
            "week",
            pd.to_datetime(["2014-05-18", "2016-10-30"]),
            np.array([228, 356]),
        ),
        (
            "m3_monthly_data",
            "date",
            pd.to_datetime(["1990-03-01", "1991-03-01"]),
            np.array([2, 14]),
        ),
        (
            "ca_hourly_electricity_data",
            "Dates",
            pd.to_datetime(["2018-01-01 02:00:00", "2018-01-05 05:00:00"]),
            np.array([2, 101]),
        ),
    ],
)
def test_dates_args(dataset, date_col, knot_dates, knot_idx, request):
    df = request.getfixturevalue(dataset)
    date_array = df[date_col]
    knot_idx2 = get_knot_idx(date_array=date_array, knot_dates=knot_dates)
    freq = pd.infer_freq(date_array)
    knot_dates2 = get_knot_dates(date_array[0], knot_idx, freq)

    assert np.all(knot_idx2 == knot_idx)
    assert np.all(knot_dates2 == knot_dates)



import numpy as np
import torch

import pyro
import pyro.distributions as dist


# FIXME: this is sort of dangerous; consider better implementation later
torch.set_default_tensor_type("torch.DoubleTensor")
pyro.enable_validation(True)


class Model:
    max_plate_nesting = 1

    def __init__(self, data):
        for key, value in data.items():
            key = key.lower()
            if isinstance(value, (list, np.ndarray)):
                # TODO: this is hackk way; please fix this later
                if key in ["which_valid_res"]:
                    # to use as index, tensor type has to be long or int
                    value = torch.tensor(value)
                elif key in ["coef_prior_list"]:
                    pass
                else:
                    # loc/scale cannot be in long format
                    # sometimes they may be supplied as int, so dtype conversion is needed
                    value = torch.tensor(value, dtype=torch.double)
            self.__dict__[key] = value

    def __call__(self):
        """
        Notes
        -----
        Labeling system:
        1. for kernel level of parameters such as rho, span, nkots, kerenel etc.,
        use suffix _lev and _coef for levels and regression to partition
        2. for knots level of parameters such as coef, loc and scale priors,
        use prefix _lev and _rr _pr for levels, regular and positive regressors to partition
        3. reduce ambigious by replacing all greeks by labels more intuitive
        use _coef, _weight etc. instead of _beta, use _scale instead of _sigma
        """

        response = self.response
        which_valid = self.which_valid_res

        n_obs = self.num_of_obs
        # n_valid = self.n_valid_res
        sdy = self.response_sd
        meany = self.mean_y
        dof = self.dof
        lev_knot_loc = self.lev_knot_loc
        seas_term = self.seas_term
        # added for tempured sampling
        T = self.t_star

        pr = self.pr
        nr = self.nr
        rr = self.rr
        n_pr = self.n_pr
        n_rr = self.n_rr
        n_nr = self.n_nr

        k_lev = self.k_lev
        k_coef = self.k_coef
        n_knots_lev = self.n_knots_lev
        n_knots_coef = self.n_knots_coef

        lev_knot_scale = self.lev_knot_scale

        resid_scale_ub = self.resid_scale_ub
        if resid_scale_ub > sdy:
            resid_scale_ub = sdy

        # expand dim to n_rr x n_knots_coef
        rr_init_knot_loc = self.rr_init_knot_loc
        rr_init_knot_scale = self.rr_init_knot_scale
        rr_knot_scale = self.rr_knot_scale

        # this does not need to expand dim since it is used as latent grand mean
        pr_init_knot_loc = self.pr_init_knot_loc
        pr_init_knot_scale = self.pr_init_knot_scale
        pr_knot_scale = self.pr_knot_scale
        nr_init_knot_loc = self.nr_init_knot_loc
        nr_init_knot_scale = self.nr_init_knot_scale
        nr_knot_scale = self.nr_knot_scale

        # prepare regressor matrix
        if n_pr == 0:
            pr = torch.zeros(0)
        if n_nr == 0:
            nr = torch.zeros(0)
        if n_rr == 0:
            rr = torch.zeros(0)
        regressors = torch.cat([rr, pr, nr], dim=-1)
        if n_pr == 0 and n_nr == 0 and n_rr == 0:
            regressors = torch.zeros(n_obs)

        response_tran = response - meany - seas_term

        # sampling begins here
        extra_out = {}

        # levels sampling
        lev_knot_tran = pyro.sample(
            "lev_knot_tran",
            dist.Normal(lev_knot_loc - meany, lev_knot_scale)
            .expand([n_knots_lev])
            .to_event(1),
        )
        lev = lev_knot_tran @ k_lev.transpose(-2, -1)
        # regular regressor sampling
        if n_rr > 0:
            # pooling latent variables
            rr_init_knot = pyro.sample(
                "rr_init_knot",
                dist.Normal(rr_init_knot_loc, rr_init_knot_scale).to_event(1),
            )
            rr_knot = pyro.sample(
                "rr_knot",
                dist.Normal(
                    rr_init_knot.unsqueeze(-1) * torch.ones(n_rr, n_knots_coef),
                    rr_knot_scale,
                ).to_event(2),
            )
            rr_coef = (rr_knot @ k_coef.transpose(-2, -1)).transpose(-2, -1)

        # positive regressor sampling
        if n_pr > 0:
            # pooling latent variables
            pr_init_knot = pyro.sample(
                "pr_knot_loc",
                dist.FoldedDistribution(
                    dist.Normal(pr_init_knot_loc, pr_init_knot_scale)
                ).to_event(1),
            )

            pr_knot = pyro.sample(
                "pr_knot",
                dist.FoldedDistribution(
                    dist.Normal(
                        pr_init_knot.unsqueeze(-1) * torch.ones(n_pr, n_knots_coef),
                        pr_knot_scale,
                    )
                ).to_event(2),
            )
            pr_coef = (pr_knot @ k_coef.transpose(-2, -1)).transpose(-2, -1)

        # negative regressor sampling
        if n_nr > 0:
            # pooling latent variables
            nr_init_knot = -1.0 * pyro.sample(
                "nr_knot_loc",
                dist.FoldedDistribution(
                    dist.Normal(nr_init_knot_loc, nr_init_knot_scale)
                ).to_event(1),
            )

            nr_knot = -1.0 * pyro.sample(
                "nr_knot",
                dist.FoldedDistribution(
                    dist.Normal(
                        nr_init_knot.unsqueeze(-1) * torch.ones(n_nr, n_knots_coef),
                        nr_knot_scale,
                    )
                ).to_event(2),
            )
            nr_coef = (nr_knot @ k_coef.transpose(-2, -1)).transpose(-2, -1)

        if n_pr == 0:
            pr_init_knot = torch.zeros(0)
            pr_knot = torch.zeros(0)
            pr_coef = torch.zeros(0)
        if n_nr == 0:
            nr_init_knot = torch.zeros(0)
            nr_knot = torch.zeros(0)
            nr_coef = torch.zeros(0)
        if n_rr == 0:
            rr_init_knot = torch.zeros(0)
            rr_knot = torch.zeros(0)
            rr_coef = torch.zeros(0)
        coef_init_knot = torch.cat([rr_init_knot, pr_init_knot, nr_init_knot], dim=-1)
        coef_knot = torch.cat([rr_knot, pr_knot, nr_knot], dim=-2)
        coef = torch.cat([rr_coef, pr_coef, nr_coef], dim=-1)
        if n_pr == 0 and n_nr == 0 and n_rr == 0:
            # coef_init_knot = torch.zeros(n_rr + n_pr + n_nr)
            # coef_knot = torch.zeros((n_rr + n_pr + n_nr, n_knots_coef))
            coef = torch.zeros(n_obs)

        # coefficients likelihood/priors
        coef_prior_list = self.coef_prior_list
        if coef_prior_list:
            for x in coef_prior_list:
                name = x["name"]
                # TODO: we can move torch conversion to init to enhance speed
                m = torch.tensor(x["prior_mean"])
                sd = torch.tensor(x["prior_sd"])
                # tp = torch.tensor(x['prior_tp_idx'])
                # idx = torch.tensor(x['prior_regressor_col_idx'])
                start_tp_idx = x["prior_start_tp_idx"]
                end_tp_idx = x["prior_end_tp_idx"]
                idx = x["prior_regressor_col_idx"]
                pyro.sample(
                    "prior_{}".format(name),
                    dist.Normal(m, sd).to_event(2),
                    obs=coef[..., start_tp_idx:end_tp_idx, idx],
                )

        # observation likelihood
        yhat = lev + (regressors * coef).sum(-1)
        # set lower and upper bound of scale parameter
        # Beta(5, 1) set up some gravity to ask for extra evidence to reduce the scale sharply
        obs_scale_base = pyro.sample("obs_scale_base", dist.Beta(5, 1)).unsqueeze(-1)
        obs_scale = obs_scale_base * resid_scale_ub

        # this line addes a tempurature to the obs fit
        with pyro.poutine.scale(scale=1.0 / T):
            pyro.sample(
                "response",
                dist.StudentT(dof, yhat[..., which_valid], obs_scale).to_event(1),
                obs=response_tran[which_valid],
            )

        log_prob = dist.StudentT(dof, yhat[..., which_valid], obs_scale).log_prob(
            response_tran[which_valid]
        )

        lev_knot = lev_knot_tran + meany

        extra_out.update(
            {
                "yhat": yhat + seas_term + meany,
                "lev": lev + meany,
                "lev_knot": lev_knot,
                "coef": coef,
                "coef_knot": coef_knot,
                "coef_init_knot": coef_init_knot,
                "obs_scale": obs_scale,
                "log_prob": log_prob,
            }
        )
        return extra_out



from copy import deepcopy
import numpy as np
import pandas as pd
import warnings
from enum import Enum

from typing import Dict, Any

from ..exceptions import ForecasterException, AbstractMethodException, IllegalArgument
from ..utils.general import is_ordered_datetime, is_even_gap_datetime
from ..template.model_template import ModelTemplate
from ..estimators.stan_estimator import StanEstimatorMCMC
from ..constants.constants import TrainingMetaKeys, PredictionMetaKeys

COMMON_MODEL_CALLABLES = [
    "get_data_input_mapper",
    "get_fitter",
    "get_init_values",
    "get_model_name",
    "get_model_param_names",
    "get_supported_estimator_types",
    "predict",
    "set_dynamic_attributes",
    "set_init_values",
]


class Forecaster(object):
    def __init__(
        self,
        model,
        estimator_type,
        response_col="y",
        date_col="ds",
        n_bootstrap_draws=-1,
        prediction_percentiles=None,
        **kwargs,
    ):
        """Abstract class for providing template of how a forecaster works by containing a model under `ModelTemplate`

        Parameters
        ----------
        model : an instance of a concrete model
        estimator_type : class of estimator
            StanEstimatorMAP, StanEstimatorMCMC, or PyroEstimatorVI
        response_col : str
            Name of response variable column, default 'y'
        date_col : str
            Name of date variable column, default 'ds'
        n_bootstrap_draws : int
            Number of samples to bootstrap in order to generate the prediction interval. For full Bayesian and
            variational inference forecasters, samples are drawn directly from original posteriors. For point-estimated
            posteriors, it will be used to sample noise parameters.  When -1 or None supplied, full Bayesian and
            variational inference forecasters will assume number of draws equal the size of original samples while
            point-estimated posteriors will mute the draw and output prediction without interval.
        prediction_percentiles : list
            List of integers of prediction percentiles that should be returned on prediction. To avoid reporting any
            confident intervals, pass an empty list

        **kwargs:
            additional arguments passed into orbit.estimators.stan_estimator or orbit.estimators.pyro_estimator
        """

        # general fields
        if not isinstance(model, ModelTemplate):
            raise ForecasterException("Invalid class of model argument supplied.")
        self._model = model
        method_list = [
            attr
            for attr in dir(model)
            # only load public methods
            if (
                callable(getattr(model, attr))
                and not attr.startswith("__")
                and not attr.startswith("_")
                and attr not in COMMON_MODEL_CALLABLES
            )
        ]
        self.extra_methods = method_list
        self.response_col = response_col
        self.date_col = date_col
        self._validate_supported_estimator_type(estimator_type)
        self.estimator_type = estimator_type
        self.estimator = self.estimator_type(**kwargs)

        self.n_bootstrap_draws = n_bootstrap_draws
        # n_bootstrap_draws holds original input
        # _n_bootstrap_draws is validated input to handle default value

        if not self.n_bootstrap_draws:
            self._n_bootstrap_draws = -1
        else:
            self._n_bootstrap_draws = int(self.n_bootstrap_draws)

        self.prediction_percentiles = prediction_percentiles
        self._prediction_percentiles = None

        if self.prediction_percentiles is None:
            self._prediction_percentiles = [5, 95]
        else:
            self._prediction_percentiles = deepcopy(self.prediction_percentiles)

        self._prediction_percentiles += [50]  # always find median
        self._prediction_percentiles = list(
            set(self._prediction_percentiles)
        )  # avoid duplicates
        self._prediction_percentiles.sort()

        # fields from fit and predict process
        # set by ._set_training_meta() in fit process
        self._training_meta = dict()
        self._training_data_input = dict()
        # fitted results after fit process
        self._posterior_samples = dict()
        self._point_posteriors = dict()
        self._training_metrics = None

        # set by ._set_prediction_meta() in predict process
        self._prediction_meta = dict()
        self.prediction_array = None

        # determine point approximation method of posteriors
        self._point_method = None

    # TODO: theoretically, we need to shrink estimator type to be consistent with the forecaster as well
    def _validate_supported_estimator_type(self, estimator_type):
        supported_estimator_types = self._model.get_supported_estimator_types()
        if estimator_type not in supported_estimator_types:
            msg_template = "Model class: {} is incompatible with Estimator: {}.  Estimator Support: {}"
            model_class = type(self._model)
            estimator_type = estimator_type
            raise ForecasterException(
                msg_template.format(
                    model_class, estimator_type, str(supported_estimator_types)
                )
            )

    def fit(self, df, **kwargs):
        """Core process for fitting a model within a forecaster"""
        estimator = self.estimator
        model_name = self._model.get_model_name()
        df = df.copy()

        # default set and validation of input data frame
        self._validate_training_df(df)
        # extract standard training metadata
        self._set_training_meta(df)
        # customize module
        self._model.set_dynamic_attributes(
            df=df, training_meta=self.get_training_meta()
        )
        # based on the model and df, set training input
        self.set_training_data_input()
        # if model provide initial values, set it
        self._model.set_init_values()

        # estimator inputs
        data_input = self.get_training_data_input()
        init_values = self._model.get_init_values()
        model_param_names = self._model.get_model_param_names()

        # note that estimator will search for the .stan, .pyro model file based on the
        # estimator type and model_name provided
        _posterior_samples, training_metrics = estimator.fit(
            model_name=model_name,
            model_param_names=model_param_names,
            data_input=data_input,
            fitter=self._model.get_fitter(),
            init_values=init_values,
            **kwargs,
        )
        self._posterior_samples = _posterior_samples
        self._training_metrics = training_metrics

    def _set_training_meta(self, df):
        """A default pre-processing and information gathering from training input dataframe"""
        training_meta = dict()
        response = df[self.response_col].values
        training_meta[TrainingMetaKeys.RESPONSE.value] = response
        training_meta[TrainingMetaKeys.DATE_ARRAY.value] = pd.to_datetime(
            df[self.date_col]
        ).reset_index(drop=True)
        training_meta[TrainingMetaKeys.NUM_OF_OBS.value] = len(response)
        training_meta[TrainingMetaKeys.RESPONSE_SD.value] = np.nanstd(response)
        training_meta[TrainingMetaKeys.RESPONSE_MEAN.value] = np.nanmean(response)
        training_meta[TrainingMetaKeys.START.value] = df[self.date_col].iloc[0]
        training_meta[TrainingMetaKeys.END.value] = df[self.date_col].iloc[-1]
        training_meta[TrainingMetaKeys.DATE_COL.value] = self.date_col
        training_meta[TrainingMetaKeys.RESPONSE_COL.value] = self.response_col
        self._training_meta = training_meta

    def get_training_meta(self):
        return deepcopy(self._training_meta)

    def set_forecaster_training_meta(self, data_input):
        """A empty function to be derived from child to set forecaster specific meta in training

        Parameters
        ----------
        data_input : dict
            training data input / meta
        Returns
        -------
        dict : updated training data input / meta
        """

        return data_input

    def set_training_data_input(self):
        """Collects data attributes into a dict for sampling/optimization api"""
        # refresh a clean dict
        data_input_mapper = self._model.get_data_input_mapper()
        if not data_input_mapper:
            raise ForecasterException("Empty or invalid data_input_mapper")

        # always get standard input from training
        training_meta = self.get_training_meta()

        training_data_input = {
            TrainingMetaKeys.RESPONSE.value.upper(): training_meta[
                TrainingMetaKeys.RESPONSE.value
            ],
            TrainingMetaKeys.RESPONSE_SD.value.upper(): training_meta[
                TrainingMetaKeys.RESPONSE_SD.value
            ],
            TrainingMetaKeys.RESPONSE_MEAN.value.upper(): training_meta[
                TrainingMetaKeys.RESPONSE_MEAN.value
            ],
            TrainingMetaKeys.NUM_OF_OBS.value.upper(): training_meta[
                TrainingMetaKeys.NUM_OF_OBS.value
            ],
        }

        training_data_input = self.set_forecaster_training_meta(
            data_input=training_data_input
        )

        if isinstance(data_input_mapper, list):
            # if a list is provided, we assume an upper case in the mapper and reuse as the input value
            for key in data_input_mapper:
                key_lower = key.lower()
                input_value = getattr(self._model, key_lower, None)
                if input_value is None:
                    raise ForecasterException(
                        "{} is missing from data input".format(key_lower)
                    )
                # stan accepts bool as int only
                if isinstance(input_value, bool):
                    input_value = int(input_value)
                training_data_input[key] = input_value
        elif issubclass(data_input_mapper, Enum):
            # isinstance(data_input_mapper, object):
            for key in data_input_mapper:
                # mapper keys in upper case; attributes is defined in lower case; need a cae casting in conversion
                key_lower = key.name.lower()
                input_value = getattr(self._model, key_lower, None)
                if input_value is None:
                    raise ForecasterException(
                        "{} is missing from data input".format(key_lower)
                    )
                if isinstance(input_value, bool):
                    # stan accepts bool as int only
                    input_value = int(input_value)
                training_data_input[key.value] = input_value
        else:
            raise Exception(
                "Invalid type: data_input_mapper needs to be either an Enum or list."
            )

        self._training_data_input = training_data_input

    def get_training_data_input(self):
        return self._training_data_input

    def _validate_training_df(self, df):
        df_columns = df.columns

        # validate date_col
        if self.date_col not in df_columns:
            raise ForecasterException(
                "DataFrame does not contain `date_col`: {}".format(self.date_col)
            )

        # validate ordering of time series
        date_array = pd.to_datetime(df[self.date_col]).reset_index(drop=True)
        if not is_ordered_datetime(date_array):
            raise ForecasterException("Datetime index must be ordered and not repeat")

        if not is_even_gap_datetime(date_array):
            warnings.warn("Datetime index is not evenly distributed")

        # validate response variable is in df
        if self.response_col not in df_columns:
            raise ForecasterException(
                "DataFrame does not contain `response_col`: {}".format(
                    self.response_col
                )
            )

    def is_fitted(self):
        # if either point posterior or posterior_samples are non-empty, claim it as fitted model (true),
        # else false.
        if bool(self._posterior_samples):
            return True
        for key in self._point_posteriors.keys():
            if bool(self._point_posteriors[key]):
                return True
        return False

    def predict(self, df: pd.DataFrame, **kwargs) -> pd.DataFrame:
        """Predict interface requires concrete implementation from child class"""
        raise AbstractMethodException(
            "Abstract method.  Model should implement concrete .predict()."
        )

    def _set_prediction_meta(self, df: pd.DataFrame) -> pd.DataFrame:
        """A default pre-processing and information gathering from prediction input dataframe"""
        # remove reference from original input
        df = df.copy()

        # get prediction df meta
        prediction_meta = {
            PredictionMetaKeys.DATE_ARRAY.value: pd.to_datetime(
                df[self.date_col]
            ).reset_index(drop=True),
            PredictionMetaKeys.PREDICTION_DF_LEN.value: df.shape[0],
            PredictionMetaKeys.START.value: df[self.date_col].iloc[0],
            PredictionMetaKeys.END.value: df[self.date_col].iloc[-1],
        }

        if not is_ordered_datetime(prediction_meta[TrainingMetaKeys.DATE_ARRAY.value]):
            raise ForecasterException("Datetime index must be ordered and not repeat")

        if not is_even_gap_datetime(prediction_meta[TrainingMetaKeys.DATE_ARRAY.value]):
            warnings.warn("Datetime index is not evenly distributed")

        # TODO: validate that all regressor columns are present, if any

        if (
            prediction_meta[PredictionMetaKeys.START.value]
            < self._training_meta[TrainingMetaKeys.START.value]
        ):
            raise ForecasterException("Prediction start must be after training start.")

        trained_len = self._training_meta[TrainingMetaKeys.NUM_OF_OBS.value]

        # If we cannot find a match of prediction range, assume prediction starts right after train
        # end
        if (
            prediction_meta[PredictionMetaKeys.START.value]
            > self._training_meta[TrainingMetaKeys.END.value]
        ):
            # time index for prediction start and end
            start_idx = trained_len
            end_idx = start_idx + df.shape[0]
            forecast_dates = set(prediction_meta[TrainingMetaKeys.DATE_ARRAY.value])
            n_forecast_steps = len(forecast_dates)
        else:
            # # compute how many steps to forecast
            # forecast_dates = set(
            #     prediction_meta[TrainingMetaKeys.DATE_ARRAY.value]
            # ) - set(self._training_meta[TrainingMetaKeys.DATE_ARRAY.value])
            # # check if prediction df is a subset of training df
            # # e.g. "negative" forecast steps
            # n_forecast_steps = len(forecast_dates) or -(
            #     len(
            #         set(self._training_meta[TrainingMetaKeys.DATE_ARRAY.value])
            #         - set(prediction_meta[TrainingMetaKeys.DATE_ARRAY.value])
            #     )
            # )

            # time index for prediction start and end
            start_idx = pd.Index(
                self._training_meta[TrainingMetaKeys.DATE_ARRAY.value]
            ).get_loc(prediction_meta[PredictionMetaKeys.START.value])
            end_idx = start_idx + df.shape[0]
            n_forecast_steps = end_idx - trained_len

        prediction_meta.update(
            {
                PredictionMetaKeys.START_INDEX.value: start_idx,
                PredictionMetaKeys.END_INDEX.value: end_idx,
                PredictionMetaKeys.FUTURE_STEPS.value: n_forecast_steps,
            }
        )

        self._prediction_meta = prediction_meta

    def get_prediction_meta(self) -> Dict[str, Any]:
        return deepcopy(self._prediction_meta)

    def get_training_metrics(self) -> Dict[str, Any]:
        return deepcopy(self._training_metrics)

    def get_posterior_samples(self, relabel: bool = False, permute: bool = True):
        """
        Parameters
        ----------
        relabel : bool
            whether returns posteriors after relabeling some parameters such as regressors etc.
        permute : bool
            default as true where chain information will be masked and the chain dimension will be collapsed; when
        it is set to false, additional chain dimension will be introduced at front before batch / sample dimension
        this is useful for arviz plotting. This argument is only effective for stan-mcmc estimator.

        Returns
        -------
        OrderedDict
            dictionary where each item represents the posterior samples of a specific parameter.
        """
        posterior_samples = deepcopy(self._posterior_samples)
        if relabel:
            regressors = self.get_regressors()
            if len(regressors) > 0:
                if len(regressors) == 1:
                    posterior_samples[regressors[0]] = posterior_samples["beta"]
                else:
                    for i, regressor in enumerate(regressors):
                        posterior_samples[regressor] = posterior_samples["beta"][:, i]
                del posterior_samples["beta"]

        if not permute:
            if self.estimator_type == StanEstimatorMCMC:
                for key, val in posterior_samples.items():
                    posterior_samples[key] = val.reshape(
                        (
                            self.estimator.chains,
                            self.estimator._num_sample_per_chain,
                            *val.shape[1:],
                        )
                    )
        return posterior_samples

    def get_point_posteriors(self) -> Dict[str, Any]:
        return deepcopy(self._point_posteriors)

    def load_extra_methods(self):
        pass

    def get_regressors(self):
        if hasattr(self._model, "_regressor_col"):
            return deepcopy(self._model._regressor_col)
        else:
            return list()

    def make_future_df(self, periods: int = 1) -> pd.DataFrame:
        """Given an Orbit Forecaster and number of periods, return a dataframe for future prediction

        Parameters
        ----------
        self : orbit.forecaster.Forecaster
        periods : int
            number of periods to generate future data frame

        Returns
        -------
        df : future dataframe

        Notes
        -----
        Alert: Right now this only works on future dataframe that doesn't require any regressors.
        """
        if periods < 1:
            raise IllegalArgument("Periods need to be greater than or equal to 1.")
        train_meta = self.get_training_meta()
        date_array = train_meta[TrainingMetaKeys.DATE_ARRAY.value]
        date_col = train_meta[TrainingMetaKeys.DATE_COL.value]
        train_end = date_array[len(date_array) - 1]
        freq = pd.infer_freq(date_array)
        future_date_array = pd.date_range(
            start=train_end, periods=periods + 1, freq=freq
        )[1:]
        future_df = pd.DataFrame(future_date_array).rename(columns={0: date_col})

        return future_df



import json
import os
import platform

import importlib_resources
from cmdstanpy import CmdStanModel

from typing import Optional

from ..utils.logger import get_logger

logger = get_logger("orbit")
# Update this to read from orbit/config.json/ORBIT_MODELS for consistency?
MODELS = ["dlt", "ets", "lgt", "ktrlite"]

# Properly set up compiler paths again
orbit_config = importlib_resources.files("orbit") / "config.json"
IS_WINDOWS = platform.platform().startswith("Win")
if IS_WINDOWS:
    with open(orbit_config) as f:
        config = json.load(f)
    CMDSTAN_VERSION = config["CMDSTAN_VERSION"]
    paths = [
        f"~/.cmdstan/cmdstan-{CMDSTAN_VERSION}/stan/lib/stan_math/lib/tbb",
        "~/.cmdstan/RTools40/mingw64/bin",
        "~/.cmdstan/RTools40/usr/bin",
    ]
    for path_string in paths:
        # or use sys.path.append, potentially works across platforms
        os.environ["Path"] += ";" + os.path.normpath(os.path.expanduser(path_string))


def get_compiled_stan_model(
    stan_model_name: str = "",
    stan_file_path: Optional[str] = None,
    exe_file_path: Optional[str] = None,
    force_compile: bool = False,
) -> CmdStanModel:
    """Return a compiled Stan model using CmdStan.
    This includes both prepackaged models as well as user provided models through stan_file_path.

    Parameters
    ----------
    stan_model_name :
        The name of the Stan model to use. Use this for the built in models (dlt, ets, ktrlite, lgt)
    stan_file_path :
        The path to the Stan file to use. If not provided, the default is to search for the file in the 'orbit' package.
        If provided, function will ignore the stan_model_name parameter, and will compile the provide stan_file_path
        into executable in place (same folder as stan_file_path)
    exe_file_path :
        The path to the Stan-exe file to use. If not provided, the default is to search for the file
        in the 'orbit' package. If provided, function will ignore the stan_model_name parameter,
        and will compile the provide stan_file_path into executable in place (same folder as stan_file_path)
    Returns
    -------
    sm : CmdStanModel
        A compiled Stan model.
    """
    if (stan_file_path is not None) or (exe_file_path is not None):
        sm = CmdStanModel(
            stan_file=stan_file, exe_file=exe_file_path, force_compile=force_compile
        )
    else:
        # Load orbit included cmdstan models
        # Some oddities here. if not providing exe_file, CmdStanModel would delete the actual executable file.
        # This is a stop gap fix until actual cause is identified.
        stan_file = importlib_resources.files("orbit") / f"stan/{stan_model_name}.stan"
        EXTENSION = ".exe" if platform.system() == "Windows" else ""
        exe_file = (
            importlib_resources.files("orbit") / f"stan/{stan_model_name}{EXTENSION}"
        )
        # Check if exe is older than .stan file.
        # This behavior is default on CmdStanModel if we don't have to specify the exe_file.
        # if not os.path.isfile(exe_file) or (
        #     os.path.getmtime(exe_file) <= os.path.getmtime(stan_file)
        # ):

        if not os.path.isfile(exe_file) or force_compile:
            logger.info(f"Compiling stan model:{stan_file}. ETA 3 - 5 mins.")
            sm = CmdStanModel(stan_file=stan_file)
        else:
            sm = CmdStanModel(stan_file=stan_file, exe_file=exe_file)

    return sm


# TODO: Is this needed?
class suppress_stdout_stderr:
    """
    A context manager for doing a "deep suppression" of stdout and stderr in
    Python, i.e. will suppress all print, even if the print originates in a
    compiled C/Fortran sub-function.
       This will not suppress raised exceptions, since exceptions are printed
    to stderr just before a script exits, and after the context manager has
    exited (at least, I think that is why it lets exceptions through).

    """

    def __init__(self):
        # Open a pair of null files
        self.null_fds = [os.open(os.devnull, os.O_RDWR) for x in range(2)]
        # Save the actual stdout (1) and stderr (2) file descriptors.
        self.save_fds = [os.dup(1), os.dup(2)]

    def __enter__(self):
        # Assign the null pointers to stdout and stderr.
        os.dup2(self.null_fds[0], 1)
        os.dup2(self.null_fds[1], 2)

    def __exit__(self, *_):
        # Re-assign the real stdout/stderr back to (1) and (2)
        os.dup2(self.save_fds[0], 1)
        os.dup2(self.save_fds[1], 2)
        # Close the null files
        for fd in self.null_fds + self.save_fds:
            os.close(fd)



import pytest
import pandas as pd
import numpy as np

from orbit.diagnostics.backtest import TimeSeriesSplitter, BackTester
from orbit.diagnostics.metrics import smape, wmape, mae
from orbit.models import LGT, KTRLite


@pytest.mark.parametrize(
    "scheduler_args",
    [
        {"min_train_len": 100, "incremental_len": 100, "forecast_len": 20},
        {"incremental_len": 100, "forecast_len": 20, "n_splits": 3},
        {"forecast_len": 20, "n_splits": 1},
        {"forecast_len": 1, "n_splits": 3},
    ],
    ids=["use_min_train_len", "use_n_splits", "single_split", "one_forecast_len"],
)
def test_time_series_splitter(scheduler_args):
    np_array = np.random.randn(300, 4)
    df = pd.DataFrame(np_array)

    tss = TimeSeriesSplitter(
        df=df,
        **scheduler_args,
    )

    assert tss.n_splits == len(list(tss.split()))


@pytest.mark.parametrize(
    "scheduler_args",
    [
        {"min_train_len": 100, "incremental_len": 100, "forecast_len": 20},
        {"incremental_len": 100, "forecast_len": 20, "n_splits": 3},
        {"forecast_len": 20, "n_splits": 1},
        {"forecast_len": 1, "n_splits": 3},
    ],
    ids=["use_min_train_len", "use_n_splits", "single_split", "one_forecast_len"],
)
def test_backtester_scheduler_args(iclaims_training_data, scheduler_args):
    df = iclaims_training_data

    lgt = LGT(
        response_col="claims",
        date_col="week",
        seasonality=1,
        verbose=False,
        estimator="stan-map",
    )

    backtester = BackTester(
        model=lgt,
        df=df,
        **scheduler_args,
    )

    backtester.fit_predict()
    eval_out = backtester.score(metrics=[smape])
    assert np.all(eval_out["metric_values"].values > 0)


@pytest.mark.parametrize(
    "metrics",
    [None, [smape, wmape, mae], smape],
    ids=["default", "multi-metrics", "single-metric"],
)
def test_backtester_test_metrics(iclaims_training_data, metrics):
    df = iclaims_training_data

    lgt = LGT(
        response_col="claims",
        date_col="week",
        seasonality=1,
        verbose=False,
        estimator="stan-map",
    )

    backtester = BackTester(
        model=lgt,
        df=df,
        forecast_len=3,
        n_splits=1,
    )

    backtester.fit_predict()
    eval_out = backtester.score(metrics=metrics)
    evaluated_metrics = set(eval_out["metric_name"].tolist())

    if metrics is None:
        expected_metrics = [x.__name__ for x in backtester._default_metrics]
    elif isinstance(metrics, list):
        expected_metrics = [x.__name__ for x in metrics]
    else:
        expected_metrics = [metrics.__name__]

    assert set(expected_metrics) == evaluated_metrics


@pytest.mark.parametrize(
    "missing_flag", [False, True], ids=["full-values", "missing-values"]
)
@pytest.mark.parametrize(
    "make_daily_data", [({"seasonality": "single", "with_coef": False})], indirect=True
)
def test_backtester_ktr_and_missing_val(make_daily_data, missing_flag):
    train_df, test_df, _ = make_daily_data
    df = pd.concat([train_df, test_df], axis=0, ignore_index=True)
    if missing_flag:
        # create a missing value in testing
        df.loc[df.shape[0] - 3, "response"] = np.nan
        # create a missing value in training
        df.loc[10, "response"] = np.nan

    ktr = KTRLite(
        date_col="date", response_col="response", seasonality=[365.25], verbose=False
    )

    bt = BackTester(
        model=ktr,
        df=df,
        n_splits=3,
        incremental_len=100,
        forecast_len=20,
    )

    bt.fit_predict()
    predicted_df = bt.get_predicted_df()
    assert set(predicted_df["split_key"].tolist()) == {0, 1, 2}

    bt_score_df = bt.score(include_training_metrics=False)
    num_testing_metrics = 6
    expected_shape = (num_testing_metrics, 3)
    assert bt_score_df.shape == expected_shape

    testing_metrics_df = bt_score_df[~bt_score_df["is_training_metric"]]
    # rmsse is the only one not working for null values; otherwise, they should have valid values
    if missing_flag:
        metric_vals = testing_metrics_df.loc[
            testing_metrics_df["metric_name"] != "rmsse", "metric_values"
        ].values
        assert np.all(~np.isnan(metric_vals))
        missing_metric_val = testing_metrics_df.loc[
            testing_metrics_df["metric_name"] == "rmsse", "metric_values"
        ].values
        assert np.all(np.isnan(missing_metric_val))
    else:
        metric_vals = testing_metrics_df["metric_values"].values
        assert np.all(~np.isnan(metric_vals))


def test_backtester_with_training_data(iclaims_training_data):
    df = iclaims_training_data

    lgt = LGT(
        response_col="claims",
        date_col="week",
        seasonality=1,
        verbose=False,
        n_bootstrap_draws=100,
        prediction_percentiles=[10, 90],
        estimator="stan-map",
    )

    backtester = BackTester(
        model=lgt,
        df=df,
        min_train_len=100,
        incremental_len=100,
        forecast_len=20,
    )

    backtester.fit_predict()
    out = backtester.get_predicted_df()
    eval_out = backtester.score(include_training_metrics=True)

    expected_out_columns = [
        "date",
        "split_key",
        "training_data",
        "actual",
        "prediction",
        "prediction_10",
        "prediction_90",
    ]
    assert set(out.columns.tolist()) == set(expected_out_columns)

    evaluated_test_metrics = set(
        eval_out.loc[~eval_out["is_training_metric"], "metric_name"].tolist()
    )
    evaluated_train_metrics = set(
        eval_out.loc[eval_out["is_training_metric"], "metric_name"].tolist()
    )
    expected_test_metrics = [x.__name__ for x in backtester._default_metrics]
    expected_train_metrics = list(
        filter(
            lambda x: backtester._get_metric_callable_signature(x)
            == {"actual", "prediction"},
            backtester._default_metrics,
        )
    )
    expected_train_metrics = [x.__name__ for x in expected_train_metrics]

    assert set(expected_test_metrics) == evaluated_test_metrics
    assert set(expected_train_metrics) == evaluated_train_metrics

    # default metric has 6 values where rmsse is only used in test metric
    num_training_metrics = 5
    num_testing_metrics = 6

    train_metric_val = eval_out.loc[
        eval_out["is_training_metric"], "metric_values"
    ].values
    test_metric_val = eval_out.loc[
        ~eval_out["is_training_metric"], "metric_values"
    ].values

    assert len(train_metric_val) == num_training_metrics
    assert len(test_metric_val) == num_testing_metrics
    assert np.all(~np.isnan(train_metric_val))
    assert np.all(~np.isnan(test_metric_val))



import pytest
import numpy as np

from orbit.models import LGT, DLT, KTR, KTRLite


@pytest.mark.parametrize("estimator", ["stan-mcmc", "stan-map"])
def test_wbic_dlt(make_weekly_data, estimator):
    train_df, test_df, coef = make_weekly_data

    dlt = DLT(
        response_col="response",
        date_col="week",
        regressor_col=list("abcdef"),
        seasonality=52,
        num_warmup=100,
        num_sample=100,
        seed=6666,
        estimator=estimator,
    )

    if estimator == "stan-mcmc":
        wbic_val = dlt.fit_wbic(df=train_df)
        assert np.isclose(wbic_val, dlt.get_wbic())

    elif estimator == "stan-map":
        dlt.fit(df=train_df)
        dlt.get_bic()


def test_wbic_lgt(make_weekly_data):
    train_df, test_df, coef = make_weekly_data

    lgt1 = LGT(
        response_col="response",
        date_col="week",
        regressor_col=list("abcdef"),
        seasonality=52,
        num_warmup=100,
        num_sample=100,
        seed=6666,
        estimator="stan-mcmc",
    )
    wbic_val1 = lgt1.fit_wbic(df=train_df)

    lgt2 = LGT(
        response_col="response",
        date_col="week",
        regressor_col=list("abcdef"),
        seasonality=52,
        num_steps=100,
        num_sample=100,
        seed=6666,
        estimator="pyro-svi",
    )
    wbic_val2 = lgt2.fit_wbic(df=train_df)

    assert np.abs(wbic_val2 - wbic_val1) / wbic_val1 <= 0.05

    lgt3 = LGT(
        response_col="response",
        date_col="week",
        regressor_col=list("abcdef"),
        seasonality=52,
        seed=6666,
        estimator="stan-map",
    )
    lgt3.fit(df=train_df)
    bic_val = lgt3.get_bic()


def test_wbic_ktr(make_weekly_data):
    train_df, test_df, coef = make_weekly_data

    ktr = KTR(
        response_col="response",
        date_col="week",
        regressor_col=list("abcdef"),
        seasonality=52,
        num_steps=100,
        num_sample=100,
        seed=6666,
        estimator="pyro-svi",
    )

    wbic_val = ktr.fit_wbic(df=train_df)
    assert np.isclose(wbic_val, ktr.get_wbic())


def test_wbic_ktrlite(make_weekly_data):
    train_df, test_df, coef = make_weekly_data

    ktrlite = KTRLite(
        response_col="response",
        date_col="week",
        seasonality=52,
        seed=6666,
        estimator="stan-map",
    )

    ktrlite.fit(df=train_df)
    ktrlite.get_bic()



import pandas as pd
import numpy as np
from datetime import datetime


def load_iclaims(end_date="2018-06-24", transform=True):
    """Load iclaims dataset

    Returns
    -------
        pd.DataFrame

    Notes
    -----
    iclaims is a dataset containing the weekly initial claims for US unemployment benefits against a few related google
    trend queries (unemploy, filling and job)from Jan 2010 - June 2018. This aims to mimick the dataset from the paper
    Predicting the Present with Bayesian Structural Time Series by SCOTT and VARIAN (2014).
    Number of claims are obtained from [Federal Reserve Bank of St. Louis] while google queries are obtained through
    Google Trends API.

    Note that dataset is transformed by natural log before fitting in order to be fitted as a multiplicative model.

    https://fred.stlouisfed.org/series/ICNSA
    https://trends.google.com/trends/?geo=US
    https://finance.yahoo.com/
    """
    url = "https://raw.githubusercontent.com/uber/orbit/master/examples/data/iclaims_example.csv"
    df = pd.read_csv(url, parse_dates=["week"], date_format="%m/%d/%y")
    df = df[df["week"] <= end_date]

    # standardize the regressors by mean; equivalent to subtracting mean after np.log
    regressors = ["trend.unemploy", "trend.filling", "trend.job", "sp500", "vix"]

    # convert to float
    for col in regressors:
        df[col] = df[col].astype(float)

    if transform:
        # log transfer
        df[["claims"] + regressors] = df[["claims"] + regressors].apply(np.log)
        # de-mean
        df[regressors] = df[regressors] - df[regressors].apply(np.mean)

    return df


def load_m4weekly():
    """Load m4 weekly sample dataset

    Returns
    -------
        pandas DataFrame

    Notes
    -----
    https://forecasters.org/resources/time-series-data/m4-competition/
    """
    url = "https://raw.githubusercontent.com/uber/orbit/master/examples/data/m4_weekly.csv"
    df = pd.read_csv(url, parse_dates=["date"])

    return df


def load_m5daily():
    """Load m5 aggregated daily dataset

    Returns
    -------
        pandas DataFrame

    Notes
    -----
    https://www.kaggle.com/c/m5-forecasting-accuracy
    """
    url = "https://raw.githubusercontent.com/uber/orbit/master/examples/data/m5_agg_demand_full.csv"
    df = pd.read_csv(url, parse_dates=["date"])

    return df


def load_m3monthly():
    """Load m3 monthly sample dataset

    Returns
    -------
        pandas DataFrame

    Notes
    -----
    https://forecasters.org/resources/time-series-data/m3-competition/
    """
    url = "https://raw.githubusercontent.com/uber/orbit/master/examples/data/m3_monthly.csv"
    df = pd.read_csv(url, parse_dates=["date"])

    return df


def load_electricity_demand():
    """Load Turkish electricity demand daily data from 1 January 2000 to 31 December 2008.
    Returns
    -------
        pandas DataFrame

    Notes
    -----
    https://robjhyndman.com/publications/complex-seasonality/

    :return:
    """

    url = "https://robjhyndman.com/data/turkey_elec.csv"
    df = pd.read_csv(url, header=None, names=["electricity"])
    df["date"] = pd.date_range(
        start="1/1/2000",
        end="31/12/2008",
        freq="D",
    )
    # re-arrange columns
    df = df[["date", "electricity"]]

    return df


def load_air_passengers():
    """Load air passengers csv from prophet github
    Returns
    -------
        pandas DataFrame

    Notes
    -----
    https://github.com/facebook/prophet/tree/master/examples

    :return:
    """

    url = "https://raw.githubusercontent.com/facebook/prophet/master/examples/example_air_passengers.csv"
    df = pd.read_csv(url, parse_dates=["ds"])

    return df


def load_energy_hourly():
    """Load Historical hourly load data from EMS for September 2021
    Returns
    -------
        pandas DataFrame

    Notes
    -----
    http://www.caiso.com/Documents/HistoricalEMSHourlyLoadforSeptember2021.xlsx
    :return:
    """

    url = "https://raw.githubusercontent.com/uber/orbit/dev/examples/data/energy_data_hourly.csv"
    df = pd.read_csv(url)
    df["hour"] = pd.to_datetime(df["Date"] + " " + df["hour"])
    df.drop(["Date", "HR"], axis=1, inplace=True)

    return df



from ..exceptions import AbstractMethodException


class ModelTemplate(object):
    """
    Notes
    -----
    contain data structure ; specify what need to fill from abstract to turn a model concrete
    """

    # class attributes
    _data_input_mapper = None
    _model_name = None
    _fitter = None
    _supported_estimator_types = None

    def __init__(self, **kwargs):
        # set by ._set_init_values
        # this is ONLY used by stan which by default used 'random'
        self._init_values = None

        # set by _set_model_param_names()
        self._model_param_names = list()

    def predict(
        self,
        posterior_estimates,
        df,
        training_meta,
        prediction_meta,
        include_error=False,
        **kwargs,
    ):
        """Predict interface for users"""
        raise AbstractMethodException(
            "Abstract method.  Model should implement concrete .predict()."
        )

    def set_dynamic_attributes(self, df, training_meta):
        """Optional; set dynamic fitting input based on input DataFrame, rather than at object instantiation"""
        pass

    def set_init_values(self):
        """Optional; set init as a callable (for Stan ONLY)"""
        pass

    def get_init_values(self):
        return self._init_values

    def get_model_param_names(self):
        return self._model_param_names

    def get_data_input_mapper(self):
        return self._data_input_mapper

    def get_model_name(self):
        return self._model_name

    def get_fitter(self):
        return self._fitter

    def get_supported_estimator_types(self):
        return self._supported_estimator_types



import json

import cmdstanpy
import importlib_resources
from ..utils.logger import get_logger

logger = get_logger("orbit")


def set_cmdstan_path():
    with open(importlib_resources.files("orbit") / "config.json") as f:
        config = json.load(f)
    CMDSTAN_VERSION = config["CMDSTAN_VERSION"]

    local_cmdstan = (
        importlib_resources.files("orbit")
        / "stan_compiled"
        / f"cmdstan-{CMDSTAN_VERSION}"
    )
    if local_cmdstan.exists():
        cmdstanpy.set_cmdstan_path(str(local_cmdstan))
        logger.debug(
            f"Local/repackaged cmdstan exists, setting path to {str(local_cmdstan)}"
        )
        return 1
    logger.debug(
        f"Cannot find local cmdstan in {str(local_cmdstan)}, using default path at ~/.cmdstan."
    )
    return 1



import numpy as np
import pandas as pd
from scipy.stats import nct
from copy import deepcopy
import torch
from enum import Enum

from ..constants.constants import (
    DEFAULT_REGRESSOR_SIGN,
    DEFAULT_REGRESSOR_BETA,
    DEFAULT_REGRESSOR_SIGMA,
    COEFFICIENT_DF_COLS,
    PredictMethod,
    PredictionKeys,
    TrainingMetaKeys,
    PredictionMetaKeys,
)
from ..exceptions import IllegalArgument, ModelException, PredictionException
from .ets import ETSModel
from ..estimators.stan_estimator import StanEstimatorMCMC, StanEstimatorMAP


class DataInputMapper(Enum):
    """
    mapping from object input to stan file
    """

    # ---------- Seasonality ---------- #
    _SEASONALITY = "SEASONALITY"
    SEASONALITY_SD = "SEASONALITY_SD"
    _SEASONALITY_SM_INPUT = "SEA_SM_INPUT"
    # ---------- Common Local Trend ---------- #
    _LEVEL_SM_INPUT = "LEV_SM_INPUT"
    _SLOPE_SM_INPUT = "SLP_SM_INPUT"
    # ---------- Global Trend ---------- #
    _GLOBAL_TREND_OPTION = "GLOBAL_TREND_OPTION"
    GLOBAL_TREND_SIGMA_PRIOR = "GB_SIGMA_PRIOR"
    GLOBAL_CAP = "G_CAP"
    GLOBAL_FLOOR = "G_FLOOR"
    # GLOBAL_LEVEL_PRIOR = 'GL_PRIOR'
    # GLOBAL_SLOPE_PRIOR = 'GB_PRIOR'
    _TIME_DELTA = "TIME_DELTA"
    # ---------- Damped Trend ---------- #
    DAMPED_FACTOR = "DAMPED_FACTOR"
    # ----------  Noise Distribution  ---------- #
    MIN_NU = "MIN_NU"
    MAX_NU = "MAX_NU"
    CAUCHY_SD = "CAUCHY_SD"
    # ----------  Regressions ---------- #
    NUM_OF_POSITIVE_REGRESSORS = "NUM_OF_PR"
    POSITIVE_REGRESSOR_MATRIX = "PR_MAT"
    POSITIVE_REGRESSOR_BETA_PRIOR = "PR_BETA_PRIOR"
    POSITIVE_REGRESSOR_SIGMA_PRIOR = "PR_SIGMA_PRIOR"
    NUM_OF_NEGATIVE_REGRESSORS = "NUM_OF_NR"
    NEGATIVE_REGRESSOR_MATRIX = "NR_MAT"
    NEGATIVE_REGRESSOR_BETA_PRIOR = "NR_BETA_PRIOR"
    NEGATIVE_REGRESSOR_SIGMA_PRIOR = "NR_SIGMA_PRIOR"
    NUM_OF_REGULAR_REGRESSORS = "NUM_OF_RR"
    REGULAR_REGRESSOR_MATRIX = "RR_MAT"
    REGULAR_REGRESSOR_BETA_PRIOR = "RR_BETA_PRIOR"
    REGULAR_REGRESSOR_SIGMA_PRIOR = "RR_SIGMA_PRIOR"
    _REGRESSION_PENALTY = "REG_PENALTY_TYPE"
    AUTO_RIDGE_SCALE = "AUTO_RIDGE_SCALE"
    LASSO_SCALE = "LASSO_SCALE"
    # handle missing values
    IS_VALID_RESPONSE = "IS_VALID_RES"
    FORECAST_HORIZON = "FORECAST_HORIZON"


class GlobalTrendOption(Enum):
    linear = 0
    loglinear = 1
    logistic = 2
    flat = 3


class BaseSamplingParameters(Enum):
    """
    base parameters in posteriors sampling
    """

    # ---------- Common Local Trend ---------- #
    LOCAL_TREND_LEVELS = "l"
    LOCAL_TREND_SLOPES = "b"
    LEVEL_SMOOTHING_FACTOR = "lev_sm"
    SLOPE_SMOOTHING_FACTOR = "slp_sm"
    # ---------- Noise Trend ---------- #
    RESIDUAL_SIGMA = "obs_sigma"
    RESIDUAL_DEGREE_OF_FREEDOM = "nu"
    # ---------- DLT Model Specific ---------- #
    LOCAL_TREND = "lt_sum"


class GlobalTrendSamplingParameters(Enum):
    GLOBAL_TREND = "gt_sum"
    GLOBAL_TREND_SLOPE = "gb"
    GLOBAL_TREND_LEVEL = "gl"


class SeasonalitySamplingParameters(Enum):
    """
    seasonality component related parameters in posteriors sampling
    """

    SEASONALITY_LEVELS = "s"
    SEASONALITY_SMOOTHING_FACTOR = "sea_sm"


class RegressionSamplingParameters(Enum):
    """
    regression component related parameters in posteriors sampling
    """

    REGRESSION_COEFFICIENTS = "beta"


class LatentSamplingParameters(Enum):
    """
    latent variables to be sampled
    """

    REGRESSION_POSITIVE_COEFFICIENTS = "pr_beta"
    REGRESSION_NEGATIVE_COEFFICIENTS = "nr_beta"
    REGRESSION_REGULAR_COEFFICIENTS = "rr_beta"
    INITIAL_SEASONALITY = "init_sea"


class RegressionPenalty(Enum):
    fixed_ridge = 0
    lasso = 1
    auto_ridge = 2


class DLTModel(ETSModel):
    """
    Parameters
    ----------
    regressor_col : list
        Names of regressor columns, if any
    regressor_sign :  list
        list with values { '+', '-', '=' } such that
        '+' indicates regressor coefficient estimates are constrained to [0, inf).
        '-' indicates regressor coefficient estimates are constrained to (-inf, 0].
        '=' indicates regressor coefficient estimates can be any value between (-inf, inf).
        The length of `regressor_sign` must be the same length as `regressor_col`. If None,
        all elements of list will be set to '='.
    regressor_beta_prior : list
        list of prior float values for regressor coefficient betas. The length of `regressor_beta_prior`
        must be the same length as `regressor_col`. If None, use non-informative priors.
    regressor_sigma_prior : list
        list of prior float values for regressor coefficient sigmas. The length of `regressor_sigma_prior`
        must be the same length as `regressor_col`. If None, use non-informative priors.
    regression_penalty : { 'fixed_ridge', 'lasso', 'auto_ridge' }
        regression penalty method
    lasso_scale : float
        float value between [0, 1], applicable only if `regression_penalty` == 'lasso'
    auto_ridge_scale : float
        float value between [0, 1], applicable only if `regression_penalty` == 'auto_ridge'
    slope_sm_input : float
        float value between [0, 1]. A larger value puts more weight on the current slope.
        If None, the model will estimate this value.
    period : int
        Used to set `time_delta` as `1 / max(period, seasonality)`. If None and no seasonality,
        then `time_delta` == 1
    damped_factor : float
        Hyperparameter float value between [0, 1]. A smaller value further dampens the previous
        global trend value. Default, 0.8
    global_trend_option : { 'flat', 'linear', 'loglinear', 'logistic' }
        Transformation function for the shape of the forecasted global trend.
    global_trend_sigma_prior : sigma prior of the global trend; default uses 1 standard deviation of response
    """

    # data labels for sampler
    _data_input_mapper = DataInputMapper
    # used to match name of `*.stan` or `*.pyro` file to look for the model
    _model_name = "dlt"
    _supported_estimator_types = [
        StanEstimatorMAP,
        StanEstimatorMCMC,
    ]

    def __init__(
        self,
        regressor_col=None,
        regressor_sign=None,
        regressor_beta_prior=None,
        regressor_sigma_prior=None,
        regression_penalty="fixed_ridge",
        lasso_scale=0.5,
        auto_ridge_scale=0.5,
        slope_sm_input=None,
        period=1,
        damped_factor=0.8,
        global_trend_option="linear",
        global_cap=1.0,
        global_floor=0.0,
        global_trend_sigma_prior=None,
        forecast_horizon=1,
        **kwargs,
    ):
        self.damped_factor = damped_factor
        self.global_trend_option = global_trend_option
        self.period = period
        # extra parameters for residuals
        self.min_nu = 5.0
        self.max_nu = 40.0

        self.slope_sm_input = slope_sm_input

        self.regressor_col = regressor_col
        self.regressor_sign = regressor_sign
        self.regressor_beta_prior = regressor_beta_prior
        self.regressor_sigma_prior = regressor_sigma_prior
        self.regression_penalty = regression_penalty
        self.lasso_scale = lasso_scale
        self.auto_ridge_scale = auto_ridge_scale

        # init static data attributes
        # the following are set by `_set_static_attributes()`
        # global trend related attributes
        self._slope_sm_input = self.slope_sm_input

        self._global_trend_option = None
        self._time_delta = 1
        self.global_cap = global_cap
        self.global_floor = global_floor

        self.global_trend_sigma_prior = global_trend_sigma_prior
        self.forecast_horizon = forecast_horizon

        # _regressor_sign stores final values after internal process of regressor_sign
        # similar to other values start with prefix _
        self._regressor_sign = self.regressor_sign
        self._regressor_beta_prior = self.regressor_beta_prior
        self._regressor_sigma_prior = self.regressor_sigma_prior
        self._regression_penalty = None
        self._regressor_col = list()

        self.num_of_regressors = 0
        # positive regressors
        self.num_of_positive_regressors = 0
        self.positive_regressor_col = list()
        self.positive_regressor_beta_prior = list()
        self.positive_regressor_sigma_prior = list()
        # negative regressors
        self.num_of_negative_regressors = 0
        self.negative_regressor_col = list()
        self.negative_regressor_beta_prior = list()
        self.negative_regressor_sigma_prior = list()
        # regular regressors
        self.num_of_regular_regressors = 0
        self.regular_regressor_col = list()
        self.regular_regressor_beta_prior = list()
        self.regular_regressor_sigma_prior = list()

        # init dynamic data attributes
        # the following are set by `_set_dynamic_attributes()` and generally set during fit()
        self.cauchy_sd = None

        # regression data
        self.regular_regressor_matrix = None
        self.positive_regressor_matrix = None
        self.negative_regressor_matrix = None

        # order matters and super constructor called after attributes are set
        # since we override _set_static_attributes()
        super().__init__(**kwargs)

    def set_init_values(self):
        """Override from base class"""
        init_values = dict()
        if self._seasonality > 1:
            init_sea = np.clip(
                np.random.normal(loc=0, scale=0.05, size=self._seasonality - 1),
                -1.0,
                1.0,
            )
            init_values[LatentSamplingParameters.INITIAL_SEASONALITY.value] = init_sea
        if self.num_of_positive_regressors > 0:
            x = np.clip(
                np.random.normal(
                    loc=0, scale=0.1, size=self.num_of_positive_regressors
                ),
                1e-5,
                2.0,
            )
            init_values[
                LatentSamplingParameters.REGRESSION_POSITIVE_COEFFICIENTS.value
            ] = x
        if self.num_of_negative_regressors > 0:
            x = np.clip(
                np.random.normal(
                    loc=0, scale=0.1, size=self.num_of_negative_regressors
                ),
                -2.0,
                -1e-5,
            )
            init_values[
                LatentSamplingParameters.REGRESSION_NEGATIVE_COEFFICIENTS.value
            ] = x
        if self.num_of_regular_regressors > 0:
            x = np.clip(
                np.random.normal(loc=0, scale=0.1, size=self.num_of_regular_regressors),
                -2.0,
                2.0,
            )
            init_values[
                LatentSamplingParameters.REGRESSION_REGULAR_COEFFICIENTS.value
            ] = x
        self._init_values = init_values

    def _validate_global_options(self):
        if self.global_trend_option not in ["flat", "linear", "loglinear", "logistic"]:
            raise IllegalArgument(
                "{} is not one of 'flat', 'linear', 'loglinear', or 'logistic'".format(
                    self.global_trend_option
                )
            )
        if self.global_trend_option == "logistic":
            if self.global_cap <= self.global_floor:
                raise IllegalArgument(
                    "In logistic global trend option, global cap{:.5f} must be greater than global floor{:.5f}.".format(
                        self.global_cap, self.global_floor
                    )
                )

    def _validate_regression_penalties(self):
        if self.regression_penalty not in ["fixed_ridge", "lasso", "auto_ridge"]:
            raise IllegalArgument(
                "{} is not one of 'fixed_ridge', 'lasso', 'auto_ridge'".format(
                    self.regression_penalty
                )
            )

    def _set_additional_trend_attributes(self):
        """Set additional trend attributes"""
        self._validate_global_options()
        self._global_trend_option = getattr(
            GlobalTrendOption, self.global_trend_option
        ).value
        self._time_delta = 1 / max(self.period, self._seasonality, 1)

        if self.slope_sm_input is None:
            self._slope_sm_input = -1

    def _set_regression_default_attributes(self):
        """set and validate regression related default attributes."""
        ##############################
        # if no regressors, end here #
        ##############################
        if self.regressor_col is None:
            # regardless of what args are set for these, if regressor_col is None
            # these should all be empty lists
            self._regressor_sign = list()
            self._regressor_beta_prior = list()
            self._regressor_sigma_prior = list()

            return

        def _validate(regression_params, valid_length):
            for p in regression_params:
                if p is not None and len(p) != valid_length:
                    raise IllegalArgument(
                        "Wrong dimension length in Regression Param Input"
                    )

        # regressor defaults
        self.num_of_regressors = len(self.regressor_col)

        _validate(
            [
                self.regressor_sign,
                self.regressor_beta_prior,
                self.regressor_sigma_prior,
            ],
            self.num_of_regressors,
        )

        if self.regressor_sign is None:
            self._regressor_sign = [DEFAULT_REGRESSOR_SIGN] * self.num_of_regressors

        if self.regressor_beta_prior is None:
            self._regressor_beta_prior = [
                DEFAULT_REGRESSOR_BETA
            ] * self.num_of_regressors

        if self.regressor_sigma_prior is None:
            self._regressor_sigma_prior = [
                DEFAULT_REGRESSOR_SIGMA
            ] * self.num_of_regressors

    def _set_regression_penalty(self):
        """set and validate regression penalty related attributes."""
        self._validate_regression_penalties()
        self._regression_penalty = getattr(
            RegressionPenalty, self.regression_penalty
        ).value

    def _set_static_regression_attributes(self):
        """set and validate regression related attributes."""
        # if no regressors, end here
        if self.regressor_col is None:
            return

        # inside *.stan files, we need to distinguish regular, positive and negative regressors
        for index, reg_sign in enumerate(self._regressor_sign):
            if reg_sign == "+":
                self.num_of_positive_regressors += 1
                self.positive_regressor_col.append(self.regressor_col[index])
                self.positive_regressor_beta_prior.append(
                    self._regressor_beta_prior[index]
                )
                self.positive_regressor_sigma_prior.append(
                    self._regressor_sigma_prior[index]
                )
            elif reg_sign == "-":
                self.num_of_negative_regressors += 1
                self.negative_regressor_col.append(self.regressor_col[index])
                self.negative_regressor_beta_prior.append(
                    self._regressor_beta_prior[index]
                )
                self.negative_regressor_sigma_prior.append(
                    self._regressor_sigma_prior[index]
                )
            else:
                self.num_of_regular_regressors += 1
                self.regular_regressor_col.append(self.regressor_col[index])
                self.regular_regressor_beta_prior.append(
                    self._regressor_beta_prior[index]
                )
                self.regular_regressor_sigma_prior.append(
                    self._regressor_sigma_prior[index]
                )

        self._regressor_col = (
            self.positive_regressor_col
            + self.negative_regressor_col
            + self.regular_regressor_col
        )

    def _set_static_attributes(self):
        """Set attributes which are independent from data
        Notes
        -----
        Overriding :func: `~orbit.models.BaseETS._set_static_attributes`
        It sets additional required attributes related to trend and regression
        """
        super()._set_static_attributes()
        self._set_additional_trend_attributes()
        self._set_regression_default_attributes()
        self._set_regression_penalty()
        self._set_static_regression_attributes()

    def _set_model_param_names(self):
        """Set posteriors keys to extract from sampling/optimization api
        Notes
        -----
        Overriding :func: `~orbit.models.BaseETS._set_model_param_names`
        It sets additional required attributes related to trend and regression
        """
        self._model_param_names += [param.value for param in BaseSamplingParameters]

        # append seasonality param names
        if self._seasonality > 1:
            self._model_param_names += [
                param.value for param in SeasonalitySamplingParameters
            ]

        # append regressors if any
        if self.num_of_regressors > 0:
            self._model_param_names += [
                RegressionSamplingParameters.REGRESSION_COEFFICIENTS.value
            ]

        if self._global_trend_option != GlobalTrendOption.flat.value:
            self._model_param_names += [
                param.value for param in GlobalTrendSamplingParameters
            ]
        else:
            self._model_param_names += [
                GlobalTrendSamplingParameters.GLOBAL_TREND_LEVEL.value
            ]
            self._model_param_names += [
                GlobalTrendSamplingParameters.GLOBAL_TREND.value
            ]

    def _validate_training_df_with_regression(self, df):
        df_columns = df.columns
        # validate regression columns
        if self.regressor_col is not None and not set(self.regressor_col).issubset(
            df_columns
        ):
            raise ModelException(
                "DataFrame does not contain specified regressor column(s)."
            )

    def _set_regressor_matrix(self, df, num_of_observations):
        """Set regressor matrix based on the input data-frame.
        Notes
        -----
        In case of absence of regression, they will be set to np.array with dim (num_of_obs, 0) to fit Stan requirement
        """
        # init of regression matrix depends on length of response vector
        self.positive_regressor_matrix = np.zeros(
            (num_of_observations, 0), dtype=np.double
        )
        self.negative_regressor_matrix = np.zeros(
            (num_of_observations, 0), dtype=np.double
        )
        self.regular_regressor_matrix = np.zeros(
            (num_of_observations, 0), dtype=np.double
        )

        # update regression matrices
        if self.num_of_positive_regressors > 0:
            self.positive_regressor_matrix = df.filter(
                items=self.positive_regressor_col,
            ).values
            if not np.all(np.isfinite(self.positive_regressor_matrix)):
                raise ModelException(
                    "Invalid regressors values. They must be all not missing and finite."
                )

        if self.num_of_negative_regressors > 0:
            self.negative_regressor_matrix = df.filter(
                items=self.negative_regressor_col,
            ).values
            if not np.all(np.isfinite(self.negative_regressor_matrix)):
                raise ModelException(
                    "Invalid regressors values. They must be all not missing and finite."
                )

        if self.num_of_regular_regressors > 0:
            self.regular_regressor_matrix = df.filter(
                items=self.regular_regressor_col,
            ).values
            if not np.all(np.isfinite(self.regular_regressor_matrix)):
                raise ModelException(
                    "Invalid regressors values. They must be all not missing and finite."
                )

    def set_dynamic_attributes(self, df, training_meta):
        """Overriding: func: `~orbit.models.BaseETS._set_dynamic_attributes"""
        super().set_dynamic_attributes(df, training_meta)
        # scalar value is suggested by the author of Rlgt
        self.cauchy_sd = (
            max(np.abs(training_meta[TrainingMetaKeys.RESPONSE.value])) / 30.0
        )
        # self._set_global_trend_priors(training_meta)
        # extra validation and settings for regression
        self._validate_training_df_with_regression(df)
        # depends on num_of_observations
        self._set_regressor_matrix(df, training_meta[TrainingMetaKeys.NUM_OF_OBS.value])

        if self.global_trend_sigma_prior is None:
            self.global_trend_sigma_prior = training_meta[
                TrainingMetaKeys.RESPONSE_SD.value
            ]

    def predict(
        self,
        posterior_estimates,
        df,
        training_meta,
        prediction_meta,
        include_error=False,
        **kwargs,
    ):
        """Vectorized version of prediction math"""
        ################################################################
        # Prediction Attributes
        ################################################################
        # n_forecast_steps = prediction_meta[PredictionMetaKeys.FUTURE_STEPS.value]
        start = prediction_meta[PredictionMetaKeys.START_INDEX.value]
        trained_len = training_meta[TrainingMetaKeys.NUM_OF_OBS.value]
        output_len = prediction_meta[PredictionMetaKeys.PREDICTION_DF_LEN.value]
        # this is only used in trend and seasonality generation
        full_len = prediction_meta[PredictionMetaKeys.END_INDEX.value]

        ################################################################
        # Model Attributes
        ################################################################
        model = deepcopy(posterior_estimates)
        for k, v in model.items():
            model[k] = torch.from_numpy(v)

        # We can pull any arbitrary value from teh dictionary because we hold the
        # safe assumption: the length of the first dimension is always the number of samples
        # thus can be safely used to determine `num_sample`. If predict_method is anything
        # other than full, the value here should be 1
        arbitrary_posterior_value = list(model.values())[0]
        num_sample = arbitrary_posterior_value.shape[0]

        # seasonality components
        seasonality_levels = model.get(
            SeasonalitySamplingParameters.SEASONALITY_LEVELS.value
        )
        seasonality_smoothing_factor = model.get(
            SeasonalitySamplingParameters.SEASONALITY_SMOOTHING_FACTOR.value
        )

        # trend components
        slope_smoothing_factor = model.get(
            BaseSamplingParameters.SLOPE_SMOOTHING_FACTOR.value
        )
        level_smoothing_factor = model.get(
            BaseSamplingParameters.LEVEL_SMOOTHING_FACTOR.value
        )
        local_trend_levels = model.get(BaseSamplingParameters.LOCAL_TREND_LEVELS.value)
        local_trend_slopes = model.get(BaseSamplingParameters.LOCAL_TREND_SLOPES.value)
        local_trend = model.get(BaseSamplingParameters.LOCAL_TREND.value)
        residual_degree_of_freedom = model.get(
            BaseSamplingParameters.RESIDUAL_DEGREE_OF_FREEDOM.value
        )
        residual_sigma = model.get(BaseSamplingParameters.RESIDUAL_SIGMA.value)

        # set an additional attribute for damped factor when it is fixed
        # get it through user input field
        damped_factor = torch.empty(num_sample, dtype=torch.double)
        damped_factor.fill_(self.damped_factor)

        if self._global_trend_option != GlobalTrendOption.flat.value:
            global_trend_level = model.get(
                GlobalTrendSamplingParameters.GLOBAL_TREND_LEVEL.value
            ).view(
                num_sample,
            )
            global_trend_slope = model.get(
                GlobalTrendSamplingParameters.GLOBAL_TREND_SLOPE.value
            ).view(
                num_sample,
            )
            global_trend = model.get(GlobalTrendSamplingParameters.GLOBAL_TREND.value)
        else:
            global_trend_level = model.get(
                GlobalTrendSamplingParameters.GLOBAL_TREND_LEVEL.value
            ).view(
                num_sample,
            )
            global_trend = model.get(GlobalTrendSamplingParameters.GLOBAL_TREND.value)

        # regression components
        regressor_beta = model.get(
            RegressionSamplingParameters.REGRESSION_COEFFICIENTS.value
        )

        ################################################################
        # Regression Component
        ################################################################
        # calculate regression component
        if self.regressor_col is not None and len(self.regressor_col) > 0:
            regressor_matrix = df[self._regressor_col].values
            if not np.all(np.isfinite(regressor_matrix)):
                raise PredictionException(
                    "Invalid regressors values. They must be all not missing and finite."
                )
            regressor_beta = regressor_beta.t()
            if len(regressor_beta.shape) == 1:
                regressor_beta = regressor_beta.unsqueeze(0)
            regressor_torch = torch.from_numpy(regressor_matrix).double()
            regression = torch.matmul(regressor_torch, regressor_beta)
            regression = regression.t()
        else:
            # regressor is always dependent with df. hence, no need to make full size
            regression = torch.zeros((num_sample, output_len), dtype=torch.double)

        ################################################################
        # Seasonality Component
        ################################################################

        # calculate seasonality component
        if self._seasonality > 1:
            if full_len <= seasonality_levels.shape[1]:
                seasonal_component = seasonality_levels[:, :full_len]
            else:
                seasonality_forecast_length = full_len - seasonality_levels.shape[1]
                seasonality_forecast_matrix = torch.zeros(
                    (num_sample, seasonality_forecast_length), dtype=torch.double
                )
                seasonal_component = torch.cat(
                    (seasonality_levels, seasonality_forecast_matrix), dim=1
                )
        else:
            seasonal_component = torch.zeros((num_sample, full_len), dtype=torch.double)

        ################################################################
        # Trend Component
        ################################################################

        # calculate level component.
        # However, if predicted end of period > training period, update with out-of-samples forecast
        if full_len <= trained_len:
            full_local_trend = local_trend[:, :full_len]
            full_global_trend = global_trend[:, :full_len]

            # in-sample error are iids
            if include_error:
                error_value = nct.rvs(
                    df=residual_degree_of_freedom.unsqueeze(-1),
                    nc=0,
                    loc=0,
                    scale=residual_sigma.unsqueeze(-1),
                    size=(num_sample, full_len),
                )

                error_value = torch.from_numpy(
                    error_value.reshape(num_sample, full_len)
                ).double()
                full_local_trend += error_value
        else:
            trend_forecast_length = full_len - trained_len
            trend_forecast_init = torch.zeros(
                (num_sample, trend_forecast_length), dtype=torch.double
            )
            full_local_trend = local_trend[:, :full_len]
            # for convenience, we lump error on local trend since the formula would
            # yield the same as yhat + noise - global_trend - seasonality - regression
            # equivalent with local_trend + noise
            # in-sample error are iids
            if include_error:
                error_value = nct.rvs(
                    df=residual_degree_of_freedom.unsqueeze(-1),
                    nc=0,
                    loc=0,
                    scale=residual_sigma.unsqueeze(-1),
                    size=(num_sample, full_local_trend.shape[1]),
                )

                error_value = torch.from_numpy(
                    error_value.reshape(num_sample, full_local_trend.shape[1])
                ).double()
                full_local_trend += error_value

            full_local_trend = torch.cat((full_local_trend, trend_forecast_init), dim=1)
            full_global_trend = torch.cat(
                (global_trend[:, :full_len], trend_forecast_init), dim=1
            )
            last_local_trend_level = local_trend_levels[:, -1]
            last_local_trend_slope = local_trend_slopes[:, -1]

            for idx in range(trained_len, full_len):
                # based on model, split cases for trend update
                curr_local_trend = (
                    last_local_trend_level
                    + damped_factor.flatten() * last_local_trend_slope
                )
                full_local_trend[:, idx] = curr_local_trend
                # idx = time - 1
                if self.global_trend_option == GlobalTrendOption.linear.name:
                    full_global_trend[:, idx] = (
                        global_trend_level + global_trend_slope * idx * self._time_delta
                    )
                elif self.global_trend_option == GlobalTrendOption.loglinear.name:
                    full_global_trend[:, idx] = (
                        global_trend_level
                        + global_trend_slope * np.log(1 + idx * self._time_delta)
                    )
                elif self.global_trend_option == GlobalTrendOption.logistic.name:
                    full_global_trend[:, idx] = self.global_floor + (
                        self.global_cap - self.global_floor
                    ) / (
                        1
                        + torch.exp(
                            -1
                            * (
                                global_trend_level
                                + global_trend_slope * idx * self._time_delta
                            )
                        )
                    )
                elif self.global_trend_option == GlobalTrendOption.flat.name:
                    full_global_trend[:, idx] = global_trend_level

                if include_error:
                    error_value = nct.rvs(
                        df=residual_degree_of_freedom,
                        nc=0,
                        loc=0,
                        scale=residual_sigma,
                        size=num_sample,
                    )
                    error_value = torch.from_numpy(error_value).double()
                    full_local_trend[:, idx] += error_value

                # now full_local_trend contains the error term and hence we need to use
                # curr_local_trend as a proxy of previous level index
                new_local_trend_level = (
                    level_smoothing_factor * full_local_trend[:, idx]
                    + (1 - level_smoothing_factor) * curr_local_trend
                )
                last_local_trend_slope = (
                    slope_smoothing_factor
                    * (new_local_trend_level - last_local_trend_level)
                    + (1 - slope_smoothing_factor)
                    * damped_factor.flatten()
                    * last_local_trend_slope
                )

                if self._seasonality > 1 and idx + self._seasonality < full_len:
                    seasonal_component[:, idx + self._seasonality] = (
                        seasonality_smoothing_factor.flatten()
                        * (
                            full_local_trend[:, idx]
                            + seasonal_component[:, idx]
                            - new_local_trend_level
                        )
                        + (1 - seasonality_smoothing_factor.flatten())
                        * seasonal_component[:, idx]
                    )

                last_local_trend_level = new_local_trend_level

        ################################################################
        # Combine Components
        ################################################################

        # trim component with right start index
        trend_component = full_global_trend[:, start:] + full_local_trend[:, start:]
        seasonal_component = seasonal_component[:, start:]

        # sum components
        pred_array = trend_component + seasonal_component + regression

        pred_array = pred_array.numpy()
        trend_component = trend_component.numpy()
        seasonal_component = seasonal_component.numpy()
        regression = regression.numpy()

        out = {
            PredictionKeys.PREDICTION.value: pred_array,
            PredictionKeys.TREND.value: trend_component,
            PredictionKeys.SEASONALITY.value: seasonal_component,
            PredictionKeys.REGRESSION.value: regression,
        }

        return out

    def get_regression_coefs(
        self,
        training_meta,
        point_method,
        point_posteriors,
        posterior_samples,
        lower=0.05,
        upper=0.95,
    ):
        """Return DataFrame regression coefficients.
          If point_method is None when fitting, return the median of coefficients.

        Parameters
        -----------
        lower : float between (0, 1). default to be 0.05
            lower bound for the CI
        upper : float between (0, 1). default to be 0.95.
            upper bound for the CI

        Returns
        -------
        pandas data frame holding the regression coefficients
        """
        # init dataframe
        coef_df = pd.DataFrame()

        # end if no regressors
        if self.num_of_regressors == 0:
            return coef_df

        _point_method = point_method
        if point_method is None:
            _point_method = PredictMethod.MEDIAN.value

        coef = point_posteriors.get(_point_method).get(
            RegressionSamplingParameters.REGRESSION_COEFFICIENTS.value
        )

        # get column names
        pr_cols = self.positive_regressor_col
        nr_cols = self.negative_regressor_col
        rr_cols = self.regular_regressor_col

        # note ordering here is not the same as `self.regressor_cols` because regressors here are grouped by signs
        regressor_cols = pr_cols + nr_cols + rr_cols

        # same note
        regressor_signs = (
            ["Positive"] * self.num_of_positive_regressors
            + ["Negative"] * self.num_of_negative_regressors
            + ["Regular"] * self.num_of_regular_regressors
        )

        coef_df[COEFFICIENT_DF_COLS.REGRESSOR] = regressor_cols
        coef_df[COEFFICIENT_DF_COLS.REGRESSOR_SIGN] = regressor_signs
        coef_df[COEFFICIENT_DF_COLS.COEFFICIENT] = coef.flatten()

        # if we have posteriors distribution and also include ci
        if point_method in [None, PredictMethod.MEAN.value, PredictMethod.MEDIAN.value]:
            coef_samples = posterior_samples.get(
                RegressionSamplingParameters.REGRESSION_COEFFICIENTS.value
            )
            coef_lower = np.quantile(coef_samples, lower, axis=0)
            coef_upper = np.quantile(coef_samples, upper, axis=0)
            coef_df[COEFFICIENT_DF_COLS.COEFFICIENT + "_lower"] = coef_lower
            coef_df[COEFFICIENT_DF_COLS.COEFFICIENT + "_upper"] = coef_upper
            n_pos = np.apply_along_axis(lambda x: np.sum(x >= 0), 0, coef_samples)
            n_total = coef_samples.shape[0]
            coef_df[COEFFICIENT_DF_COLS.PROB_COEF_POS] = n_pos / n_total
            coef_df[COEFFICIENT_DF_COLS.PROB_COEF_NEG] = 1 - n_pos / n_total

        return coef_df



import json
import os
import platform
import shutil
from pathlib import Path

from setuptools import find_packages, setup
from setuptools.command.build_py import build_py

MODEL_SOURCE_DIR = "orbit/stan"
MODELS = ["dlt", "ets", "lgt", "ktrlite"]

DESCRIPTION = "Orbit is a package for Bayesian time series modeling and inference."
BINARIES = ["diagnose", "print", "stanc", "stansummary"]

with open("orbit/config.json") as f:
    config = json.load(f)
CMDSTAN_VERSION = config["CMDSTAN_VERSION"]
IS_WINDOWS = platform.platform().startswith("Win")


def read_long_description(filename="README.md"):
    # with open(filename) as f:
    with open("README.md", "r", encoding="utf-8") as f:
        return f.read().strip()


def requirements(filename="requirements.txt"):
    with open(filename) as f:
        return f.readlines()


def remove_older_cmdstan(cmdstan_version):
    """
    Checks the default CmdStan path (~/.cmdstan) for folders with older versions
    and removes them.

    Args:
        cmdstan_version (str): The current CmdStan version.
    """
    default_path = os.path.expanduser("~/.cmdstan")

    if not os.path.exists(default_path):
        # Path doesn't exist, nothing to remove
        return

    for folder in os.listdir(default_path):
        if folder.startswith("cmdstan-"):
            # Extract version number from folder name
            folder_version = folder.split("-")[1]
            if folder_version < cmdstan_version:
                # Remove folder if version is older
                full_path = os.path.join(default_path, folder)
                try:
                    print(f"Removing older CmdStan version: {folder} : {full_path}")
                    shutil.rmtree(full_path)
                    print(f"Done.")
                except OSError as e:
                    print(f"Error removing {folder}: {e}")


def install_stan():
    """
    Compile and install stan backend
    Reference from prophet
    """
    from multiprocessing import cpu_count
    import cmdstanpy

    remove_older_cmdstan(CMDSTAN_VERSION)

    if not cmdstanpy.install_cmdstan(
        version=CMDSTAN_VERSION,
        overwrite=True,
        verbose=True,
        cores=cpu_count(),
        progress=True,
        compiler=IS_WINDOWS,
    ):
        raise RuntimeError("CmdStan failed to install in repackaged directory")

    print("Installed cmdstanpy package.")


def build_model(model: str, model_dir: str):
    import cmdstanpy

    # compile stan file in place
    model_name = f"{model}.stan"
    model_path = os.path.join(model_dir, model_name)
    print(f"Compiling stan file in place: {model_path}")
    _ = cmdstanpy.CmdStanModel(stan_file=model_path, force_compile=True)


def build_stan_models():
    for model in MODELS:
        # note: ensure copy target is a directory not a file.
        build_model(
            model=model,
            model_dir=MODEL_SOURCE_DIR,
        )


class BuildPyCommand(build_py):
    """Custom build command to make sure install cmdstanpy properly."""

    def run(self):
        print("Running build py command.")
        if not self.dry_run:
            print("Not a dry run, run with build")
            # build_stan_model(target_dir)

            # install cmdstan and compilers, in the default directory for simplicity
            install_stan()

            # build all stan models
            build_stan_models()

        print("Dry run.")
        build_py.run(self)


about = {}
here = Path(__file__).parent.resolve()
with open(here / "orbit" / "__version__.py", "r") as f:
    exec(f.read(), about)


setup(
    version=about["__version__"],
    packages=find_packages(),
    name="orbit-ml",
    description=DESCRIPTION,
    include_package_data=True,
    install_requires=requirements("requirements.txt"),
    tests_require=requirements("requirements-test.txt"),
    cmdclass={
        "build_py": BuildPyCommand,
    },
    test_suite="orbit.tests",
    long_description=read_long_description(),
    long_description_content_type="text/markdown",
    url="https://orbit-ml.readthedocs.io/en/stable/",
    zip_safe=False,
)



import logging
import multiprocessing
from abc import abstractmethod
from copy import copy
from sys import platform, version_info

from ..exceptions import EstimatorException
from ..utils.general import update_dict
from ..utils.logger import get_logger
from ..utils.set_cmdstan_path import set_cmdstan_path
from ..utils.stan import get_compiled_stan_model
from .base_estimator import BaseEstimator

logger = get_logger("orbit")

# Make sure models are using the right cmdstan folder
set_cmdstan_path()


class StanEstimator(BaseEstimator):
    """Abstract StanEstimator with shared args for all StanEstimator child classes

    Parameters
    ----------
    num_warmup : int
        Number of samples to warm up and to be discarded, default 900
    num_sample : int
        Number of samples to return, default 100
    chains : int
        Number of chains in stan sampler, default 4
    cores : int
        Number of cores for parallel processing, default max(cores, multiprocessing.cpu_count())
    algorithm : str
        If None, default to Stan defaults
    suppress_stan_log : bool
        If False, turn off cmdstanpy logger. Default as False.
    kwargs
        Additional `BaseEstimator` class args

    """

    def __init__(
        self,
        num_warmup=900,
        num_sample=100,
        chains=4,
        cores=8,
        algorithm=None,
        suppress_stan_log=True,
        **kwargs,
    ):
        # see https://mc-stan.org/cmdstanpy/users-guide/outputs.html for details
        cmdstanpy_logger = logging.getLogger("cmdstanpy")
        cmdstanpy_logger.disabled = suppress_stan_log

        super().__init__(**kwargs)
        self.num_warmup = num_warmup
        self.num_sample = num_sample
        self.chains = chains
        self.cores = cores
        self.algorithm = algorithm

        # stan_init fallback if not provided in model
        # this arg is passed in through `model_payload` in `fit()` to override
        # self.stan_init = "random"

        # init computed configs
        self._num_warmup_per_chain = None
        self._num_sample_per_chain = None
        self._num_iter_per_chain = None
        self._total_iter = None

        self._set_computed_stan_configs()

    def _set_computed_stan_configs(self):
        """Sets sampler configs based on init class attributes"""
        # make sure cores can only be as large as the device support
        self.cores = min(self.cores, multiprocessing.cpu_count())
        self._num_warmup_per_chain = int(self.num_warmup / self.chains)
        self._num_sample_per_chain = int(self.num_sample / self.chains)
        self._num_iter_per_chain = (
            self._num_warmup_per_chain + self._num_sample_per_chain
        )
        self._total_iter = self._num_iter_per_chain * self.chains

    @abstractmethod
    def fit(
        self, model_name, model_param_names, data_input, fitter=None, init_values=None
    ):
        raise NotImplementedError("Concrete fit() method must be implemented")


class StanEstimatorMCMC(StanEstimator):
    """Stan Estimator for MCMC Sampling

    Parameters
    ----------
    stan_mcmc_args : dict
        Supplemental stan mcmc args to pass to CmdStandPy.sampling()

    """

    def __init__(self, stan_mcmc_args=None, **kwargs):
        super().__init__(**kwargs)
        self.stan_mcmc_args = stan_mcmc_args

        # init computed args
        self._stan_mcmc_args = copy(self.stan_mcmc_args)
        self._set_computed_stan_mcmc_configs()

    def _set_computed_stan_mcmc_configs(self):
        self._stan_mcmc_args = update_dict({}, self._stan_mcmc_args)

    def fit(
        self,
        model_name,
        model_param_names,
        sampling_temperature,
        data_input,
        fitter=None,
        init_values=None,
    ):
        # T_STAR is used as sampling temperature which is used for WBIC calculation
        data_input.update({"T_STAR": sampling_temperature})
        if self.verbose:
            msg_template = (
                "Sampling (CmdStanPy) with chains: {:d}, cores: {:d}, temperature: {:.3f}, "
                "warmups (per chain): {:d} and samples(per chain): {:d}."
            )
            msg = msg_template.format(
                self.chains,
                self.cores,
                sampling_temperature,
                self._num_warmup_per_chain,
                self._num_sample_per_chain,
            )
            logger.info(msg)

        compiled_mod = get_compiled_stan_model(stan_model_name=model_name)

        # check https://mc-stan.org/cmdstanpy/api.html#cmdstanpy.CmdStanModel.sample
        # for additional args
        stan_mcmc_fit = compiled_mod.sample(
            data=data_input,
            iter_sampling=self._num_sample_per_chain,
            iter_warmup=self._num_warmup_per_chain,
            chains=self.chains,
            parallel_chains=self.cores,
            inits=init_values,
            seed=self.seed,
            **self._stan_mcmc_args,
        )

        stan_extract = stan_mcmc_fit.stan_variables()
        posteriors = {
            param: stan_extract[param] for param in model_param_names + ["loglk"]
        }

        training_metrics = {
            "loglk": posteriors["loglk"],
            "sampling_temperature": sampling_temperature,
        }

        return posteriors, training_metrics


class StanEstimatorMAP(StanEstimator):
    """Stan Estimator for MAP Posteriors"""

    def __init__(self, stan_map_args=None, **kwargs):
        super().__init__(**kwargs)
        self.stan_map_args = stan_map_args

        # init computed args
        self._stan_map_args = copy(self.stan_map_args)

        # set defaults
        self._set_computed_stan_map_configs()

    def _set_computed_stan_map_configs(self):
        default_stan_map_args = {}
        self._stan_map_args = update_dict(default_stan_map_args, self._stan_map_args)
        # TODO: should it be within the fit block?
        if self.verbose:
            msg_template = "Optimizing (CmdStanPy) with algorithm: {}."
            if self.algorithm is None:
                algorithm = "LBFGS"
            else:
                algorithm = self.algorithm
            msg = msg_template.format(algorithm)
            logger.info(msg)

    def fit(
        self,
        model_name,
        model_param_names,
        data_input,
        fitter=None,
        init_values=None,
    ):
        compiled_mod = get_compiled_stan_model(stan_model_name=model_name)
        data_input.update({"T_STAR": 1.0})

        # passing callable from the model as seen in `initfun1()`
        # init_values = init_values or self.stan_init

        # in case optimizing fails with given algorithm fallback to `Newton`
        # init values interface can be referred here: https://cmdstanpy.readthedocs.io/en/stable-0.9.65/api.html
        # Dict [Str, np.array] where key is the param name and value array dim matches the param dim
        try:
            stan_fit = compiled_mod.optimize(
                data=data_input,
                inits=init_values,
                seed=self.seed,
                algorithm=self.algorithm,
                **self._stan_map_args,
            )
        except RuntimeError:
            self.algorithm = "Newton"
            stan_fit = compiled_mod.optimize(
                data=data_input,
                inits=init_values,
                seed=self.seed,
                algorithm=self.algorithm,
                **self._stan_map_args,
            )

        stan_extract = stan_fit.stan_variables()
        # make sure that model param names are a subset of stan extract keys
        invalid_model_param = set(model_param_names) - set(list(stan_extract.keys()))
        if invalid_model_param:
            raise EstimatorException(
                "Stan model definition does not contain required parameters"
            )

        # `stan.optimizing` automatically returns all defined parameters
        # filter out unnecessary keys
        posteriors = {
            param: stan_extract[param] for param in model_param_names + ["loglk"]
        }

        training_metrics = {
            # loglk is needed for BIC calculation
            "loglk": stan_extract["loglk"],
            # TODO: this needs to be the full length of all parameters instead of the one we sampled?
            # TODO: i.e. should it include latent variables / derive variables?
            "num_of_params": len(model_param_names),
        }

        return posteriors, training_metrics



import logging
from typing import Optional


def make_info_logger(name: str, path: Optional[str] = None) -> logging.Logger:
    """generate new logger in a standardized way
    Returns:
        logging.Logger: logger created in INFO level and time-stamp
    """
    logger = logging.getLogger(name)
    logger.setLevel(logging.INFO)
    formatter = logging.Formatter(
        "%(asctime)s - %(name)s - %(levelname)s - %(message)s",
        datefmt="%Y-%m-%d %H:%M:%S",
    )

    if path is None:
        ch = logging.StreamHandler()
        ch.setLevel(logging.INFO)
        ch.setFormatter(formatter)
        logger.addHandler(ch)
    else:
        fh = logging.FileHandler(path, mode="w")
        fh.setLevel(logging.INFO)
        fh.setFormatter(formatter)
        logger.addHandler(fh)

    return logger


def get_logger(name: str) -> logging.Logger:
    logger = logging.getLogger(name=name)
    if len(logger.handlers) == 0:
        logger = make_info_logger(name)
    return logger
